from __future__ import annotations

import itertools
import logging
from typing import TYPE_CHECKING, cast

import pytest
from pytest_bdd import given, parsers, scenarios, then, when
from typing_extensions import TypeAlias

import cartographer.probe.scan_mode as scan_mode
import cartographer.probe.touch_mode as touch_mode
from cartographer.printer_interface import Sample
from cartographer.probe import Probe

if TYPE_CHECKING:
    from unittest.mock import Mock

    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture

    from cartographer.printer_interface import MacroParams, Toolhead


scenarios("../../features/probe.feature")

ScanMode: TypeAlias = scan_mode.ScanMode[object, Sample]
TouchMode: TypeAlias = touch_mode.TouchMode[object]


@pytest.fixture
def scan(mocker: MockerFixture):
    return mocker.MagicMock(spec=ScanMode, instance=True, autospec=True)


@pytest.fixture
def touch(mocker: MockerFixture):
    return mocker.MagicMock(spec=TouchMode, instance=True, autospec=True)


@given("a probe", target_fixture="probe")
def given_probe(scan: ScanMode, touch: TouchMode) -> Probe:
    return Probe(scan, touch)


@given("the probe measures:")
def given_probe_measurements(mocker: MockerFixture, scan: ScanMode, datatable: list[list[str]]):
    scan.perform_probe = mocker.Mock(side_effect=itertools.cycle(map(float, itertools.chain.from_iterable(datatable))))


@given("the probe is triggered")
def given_probe_triggered(mocker: MockerFixture, scan: ScanMode):
    scan.query_is_triggered = mocker.Mock(return_value=True)


@given("the probe is not triggered")
def given_probe_not_triggered(mocker: MockerFixture, scan: ScanMode):
    scan.query_is_triggered = mocker.Mock(return_value=False)


@when("I run the PROBE macro")
def when_run_probe_macro(params: MacroParams, caplog: LogCaptureFixture, probe: Probe):
    from cartographer.macros.probe import ProbeMacro

    macro = ProbeMacro(probe)
    with caplog.at_level(logging.INFO):
        macro.run(params)


@when("I run the PROBE_ACCURACY macro")
def when_run_probe_accuracy_macro(params: MacroParams, caplog: LogCaptureFixture, probe: Probe, toolhead: Toolhead):
    from cartographer.macros.probe import ProbeAccuracyMacro

    macro = ProbeAccuracyMacro(probe, toolhead)
    with caplog.at_level(logging.INFO):
        macro.run(params)


@when("I run the QUERY_PROBE macro")
def when_run_query_probe_macro(params: MacroParams, caplog: LogCaptureFixture, probe: Probe):
    from cartographer.macros.probe import QueryProbeMacro

    macro = QueryProbeMacro(probe)
    with caplog.at_level(logging.INFO):
        macro.run(params)


@then(parsers.parse("it should probe {count:d} times"))
def then_probe_count(scan: ScanMode, count: int):
    assert cast("Mock", scan.perform_probe).call_count == count



from __future__ import annotations

import multiprocessing
import traceback
from typing import TYPE_CHECKING, Callable, TypeVar, final

from typing_extensions import ParamSpec, override

from cartographer.interfaces import TaskExecutor

if TYPE_CHECKING:
    from multiprocessing.connection import Connection

    from reactor import Reactor

P = ParamSpec("P")
R = TypeVar("R")

WAIT_TIME = 0.1


@final
class KlipperMultiprocessingExecutor(TaskExecutor):
    def __init__(self, reactor: Reactor) -> None:
        self._reactor = reactor

    @override
    def run(self, fn: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
        def worker(child_conn: Connection):
            try:
                result = fn(*args, **kwargs)
                child_conn.send((False, result))
            except Exception:
                child_conn.send((True, traceback.format_exc()))
            finally:
                child_conn.close()

        parent_conn, child_conn = multiprocessing.Pipe()
        proc = multiprocessing.Process(target=worker, args=(child_conn,), daemon=True)
        proc.start()

        eventtime = self._reactor.monotonic()
        while proc.is_alive():
            eventtime = self._reactor.pause(eventtime + WAIT_TIME)

        proc.join()
        is_err, result = parent_conn.recv()
        parent_conn.close()

        if is_err:
            raise RuntimeError("Subprocess failed:\n" + result)
        return result



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import MagicMock

import pytest
from typing_extensions import override

from cartographer.klipper.endstop import KlipperEndstop
from cartographer.printer_interface import Endstop, HomingState

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from cartographer.klipper.mcu import KlipperCartographerMcu

ReactorCompletion = MagicMock


class MockEndstop(Endstop[ReactorCompletion]):
    @override
    def query_is_triggered(self, print_time: float) -> bool:
        return False

    @override
    def home_start(self, print_time: float) -> ReactorCompletion:
        return MagicMock()

    @override
    def home_wait(self, home_end_time: float) -> float:
        return home_end_time

    @override
    def on_home_end(self, homing_state: HomingState) -> None:
        return

    @override
    def get_endstop_position(self) -> float:
        return 2.0


@pytest.fixture
def mcu(mocker: MockerFixture) -> KlipperCartographerMcu:
    return mocker.MagicMock()


@pytest.fixture
def endstop() -> Endstop[object]:
    return MockEndstop()


def test_endstop_is_memoized(mcu: KlipperCartographerMcu, endstop: Endstop[ReactorCompletion]):
    a = KlipperEndstop(mcu, endstop)
    b = KlipperEndstop(mcu, endstop)

    assert a == b


def test_endstop_creates_multiple(mcu: KlipperCartographerMcu, endstop: Endstop[ReactorCompletion]):
    another_endstop = MockEndstop()
    a = KlipperEndstop(mcu, endstop)
    b = KlipperEndstop(mcu, another_endstop)

    assert a != b



from __future__ import annotations

import logging
from dataclasses import dataclass
from itertools import combinations
from typing import TYPE_CHECKING, Protocol, final

import numpy as np
from typing_extensions import override

from cartographer.lib.statistics import compute_mad
from cartographer.printer_interface import C, Endstop, HomingState, Mcu, Position, ProbeMode, S, Toolhead

if TYPE_CHECKING:
    from collections.abc import Sequence

    from cartographer.configuration import TouchModelConfiguration

logger = logging.getLogger(__name__)


STD_TOLERANCE = 0.008
RETRACT_DISTANCE = 2.0
MAX_TOUCH_TEMPERATURE = 155


class Configuration(Protocol):
    move_speed: float

    touch_samples: int
    touch_max_samples: int

    x_offset: float
    y_offset: float
    mesh_min: tuple[float, float]
    mesh_max: tuple[float, float]


class TouchError(RuntimeError):
    pass


@dataclass(frozen=True)
class TouchBoundaries:
    min_x: float
    max_x: float
    min_y: float
    max_y: float

    def is_within(self, *, x: float, y: float) -> bool:
        epsilon = 0.01
        in_x_bounds = (self.min_x - epsilon) <= x <= (self.max_x + epsilon)
        in_y_bounds = (self.min_y - epsilon) <= y <= (self.max_y + epsilon)
        return in_x_bounds and in_y_bounds

    @staticmethod
    def from_config(config: Configuration) -> TouchBoundaries:
        mesh_min_x, mesh_min_y = config.mesh_min
        mesh_max_x, mesh_max_y = config.mesh_max
        x_offset = config.x_offset
        y_offset = config.y_offset

        # For negative offsets: increase min bounds, leave max bounds unchanged
        min_x = mesh_min_x - min(x_offset, 0)  # Only subtract negative offsets
        min_y = mesh_min_y - min(y_offset, 0)  # Only subtract negative offsets
        # For positive offsets: reduce max bounds, leave min bounds unchanged
        max_x = mesh_max_x - max(x_offset, 0)  # Only subtract positive offsets
        max_y = mesh_max_y - max(y_offset, 0)  # Only subtract positive offsets

        return TouchBoundaries(
            min_x=min_x,
            max_x=max_x,
            min_y=min_y,
            max_y=max_y,
        )


@final
class TouchMode(ProbeMode, Endstop[C]):
    """Implementation for Survey Touch."""

    def get_model(self) -> TouchModelConfiguration:
        if self.model is None:
            msg = "no touch model loaded"
            raise RuntimeError(msg)
        return self.model

    @property
    @override
    def offset(self) -> Position:
        z_offset = self.model.z_offset if self.model else 0.0
        return Position(0.0, 0.0, z_offset)

    @override
    def save_z_offset(self, new_offset: float) -> None:
        self.get_model().save_z_offset(new_offset)

    @property
    @override
    def is_ready(self) -> bool:
        return self.model is not None

    def __init__(
        self,
        mcu: Mcu[C, S],
        toolhead: Toolhead,
        config: Configuration,
        *,
        model: TouchModelConfiguration | None,
    ) -> None:
        self._toolhead = toolhead
        self._mcu = mcu
        self.config = config
        self.model = model

        self.boundaries = TouchBoundaries.from_config(config)

    @override
    def perform_probe(self) -> float:
        if not self._toolhead.is_homed("z"):
            msg = "z axis must be homed before probing"
            raise RuntimeError(msg)

        if self._toolhead.get_position().z < 5:
            self._toolhead.move(z=5, speed=self.config.move_speed)
        self._toolhead.wait_moves()

        return self._run_probe()

    def _run_probe(self) -> float:
        collected: list[float] = []
        touch_samples = self.config.touch_samples
        touch_max_samples = self.config.touch_max_samples
        logger.debug("Starting touch sequence for %d samples within %d touches...", touch_samples, touch_max_samples)

        for i in range(touch_max_samples):
            trigger_pos = self.perform_single_probe()
            collected.append(trigger_pos)
            logger.debug("Touch %d: %.6f", i + 1, trigger_pos)

            if len(collected) < touch_samples:
                continue

            valid_combo = self._find_valid_combination(collected, touch_samples)
            if valid_combo is None:
                continue

            self._log_sample_stats("Acceptable touch combination found", valid_combo)

            return float(np.median(valid_combo) if len(valid_combo) > 3 else np.mean(valid_combo))

        self._log_sample_stats("No valid touch combination found in samples", collected)
        msg = f"unable to find {touch_samples} samples within tolerance after {touch_max_samples} touches"
        raise TouchError(msg)

    def _find_valid_combination(self, samples: list[float], size: int) -> tuple[float, ...] | None:
        for combo in combinations(samples, size):
            if np.std(combo) <= STD_TOLERANCE:
                return combo
        return None

    def perform_single_probe(self) -> float:
        model = self.get_model()
        if self._toolhead.get_position().z < RETRACT_DISTANCE:
            self._toolhead.move(z=RETRACT_DISTANCE, speed=self.config.move_speed)
        self._toolhead.wait_moves()
        trigger_pos = self._toolhead.z_homing_move(self, bottom=-2.0, speed=model.speed)
        pos = self._toolhead.get_position()
        self._toolhead.move(
            z=max(pos.z + RETRACT_DISTANCE, RETRACT_DISTANCE),
            speed=self.config.move_speed,
        )
        return trigger_pos

    @override
    def home_start(self, print_time: float) -> C:
        model = self.get_model()
        if model.threshold <= 0:
            msg = "threshold must be greater than 0"
            raise RuntimeError(msg)

        pos = self._toolhead.get_position()
        if not self.is_within_boundaries(x=pos.x, y=pos.y):
            msg = f"position ({pos.x:.2f},{pos.y:.2f}) is outside of the touch boundaries"
            raise RuntimeError(msg)

        nozzle = self._toolhead.get_extruder_temperature()
        if nozzle.current > MAX_TOUCH_TEMPERATURE or nozzle.target > MAX_TOUCH_TEMPERATURE:
            msg = f"nozzle temperature must be below {MAX_TOUCH_TEMPERATURE - 5:d}C"
            raise RuntimeError(msg)
        return self._mcu.start_homing_touch(print_time, model.threshold)

    def is_within_boundaries(self, *, x: float, y: float) -> bool:
        return self.boundaries.is_within(x=x, y=y)

    @override
    def on_home_end(self, homing_state: HomingState) -> None:
        if self not in homing_state.endstops:
            return
        if not homing_state.is_homing_z():
            return

        homing_state.set_z_homed_position(self.get_model().z_offset)

    @override
    def home_wait(self, home_end_time: float) -> float:
        return self._mcu.stop_homing(home_end_time)

    @override
    def query_is_triggered(self, print_time: float) -> bool:
        # Touch endstop is never in a triggered state.
        return False

    @override
    def get_endstop_position(self) -> float:
        return self.offset.z

    def _log_sample_stats(self, message: str, samples: Sequence[float]) -> None:
        max_v, min_v = max(samples), min(samples)
        mean = np.mean(samples)
        median = np.median(samples)
        range_v = max_v - min_v
        std_dev = np.std(samples)
        mad = compute_mad(samples)
        logger.debug(
            """
                %s: (%s)
                maximum %.6f, minimum %.6f, range %.6f,
                average %.6f, median %.6f, standard deviation %.6f,
                median absolute deviation %.6f
                """,
            message,
            ", ".join(f"{s:.6f}" for s in samples),
            max_v,
            min_v,
            range_v,
            mean,
            median,
            std_dev,
            mad,
        )



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import override

from cartographer.configuration import (
    Configuration as CartographerConfiguration,
)
from cartographer.configuration import (
    ScanModelConfiguration,
    ScanModelFit,
    TouchModelConfiguration,
)

if TYPE_CHECKING:
    from configfile import ConfigWrapper


T = TypeVar("T", bound=Enum)


def get_enum_choice(config: ConfigWrapper, option: str, enum_type: type[T], default: T) -> T:
    choice = config.get(option, default.value)
    if choice not in enum_type._value2member_map_:
        msg = f"invalid choice '{choice}' for option '{option}'"
        raise config.error(msg)
    return enum_type(choice)


def get_coordinate_point(config: ConfigWrapper, option: str) -> tuple[float, float]:
    x, y = config.getfloatlist(option, count=2)
    return x, y


class KlipperCartographerConfiguration(CartographerConfiguration):
    def __init__(self, config: ConfigWrapper) -> None:
        self._config: ConfigWrapper = config
        self.x_offset: float = config.getfloat("x_offset")
        self.y_offset: float = config.getfloat("y_offset")
        self.backlash_compensation: float = config.getfloat("backlash_compensation", 0)
        self.move_speed: float = config.getfloat("move_speed", default=50, above=0)
        self.verbose: bool = config.getboolean("verbose", default=False)

        config_name = config.get_name()

        scan_config = config.getsection(f"{config_name} scan")
        # self.scan_samples: int = scan_config.getint("samples", default=20, minval=20)
        self.scan_samples: int = 20
        self.scan_mesh_runs: int = scan_config.getint("mesh_runs", default=1, minval=1)

        touch_config = config.getsection(f"{config_name} touch")
        self.touch_samples: int = touch_config.getint("samples", default=5, minval=3)
        self.touch_max_samples: int = touch_config.getint(
            "max_samples", default=self.touch_samples * 2, minval=self.touch_samples
        )

        self.scan_models: dict[str, ScanModelConfiguration] = {
            cfg.name: cfg
            for cfg in map(
                KlipperScanModelConfiguration.from_config, config.get_prefix_sections(f"{config_name} scan_model")
            )
        }
        self.touch_models: dict[str, TouchModelConfiguration] = {
            cfg.name: cfg
            for cfg in map(
                KlipperTouchModelConfiguration.from_config, config.get_prefix_sections(f"{config_name} touch_model")
            )
        }

        mesh_config = config.getsection("bed_mesh")
        self.scan_speed: float = mesh_config.getfloat("speed", default=50, minval=50)
        self.scan_height: float = mesh_config.getfloat("horizontal_move_z", default=4, minval=1)
        self.zero_reference_position: tuple[float, float] = get_coordinate_point(mesh_config, "zero_reference_position")
        self.mesh_min: tuple[float, float] = get_coordinate_point(mesh_config, "mesh_min")
        self.mesh_max: tuple[float, float] = get_coordinate_point(mesh_config, "mesh_max")

    @override
    def save_new_scan_model(self, name: str, model: ScanModelFit) -> KlipperScanModelConfiguration:
        section_name = f"{self._config.get_name()} scan_model {name}"
        configfile = self._config.get_printer().lookup_object("configfile")
        configfile.set(section_name, "z_offset", 0)
        configfile.set(section_name, "coefficients", ", ".join(map(str, model.coefficients)))
        configfile.set(section_name, "domain", ", ".join(map(str, model.domain)))

        return KlipperScanModelConfiguration(
            self._config.getsection(section_name),
            name=name,
            coefficients=model.coefficients,
            domain=model.domain,
            z_offset=0,
        )

    @override
    def save_new_touch_model(self, name: str, speed: float, threshold: int) -> TouchModelConfiguration:
        section_name = f"{self._config.get_name()} touch_model {name}"
        configfile = self._config.get_printer().lookup_object("configfile")
        configfile.set(section_name, "threshold", threshold)
        configfile.set(section_name, "speed", f"{speed:.1f}")
        configfile.set(section_name, "z_offset", -0.05)

        return KlipperTouchModelConfiguration(
            self._config.getsection(section_name),
            name=name,
            threshold=threshold,
            speed=speed,
            z_offset=0,
        )


class KlipperScanModelConfiguration(ScanModelConfiguration):
    def __init__(
        self,
        config: ConfigWrapper,
        *,
        name: str,
        coefficients: list[float],
        domain: tuple[float, float],
        z_offset: float,
    ) -> None:
        self._config: ConfigWrapper = config
        self.name: str = name
        self.coefficients: list[float] = coefficients
        self.domain: tuple[float, float] = domain
        self.z_offset: float = z_offset

    @override
    def save_z_offset(self, new_offset: float) -> None:
        self._config.get_printer().lookup_object("configfile").set(
            self._config.get_name(), "z_offset", f"{new_offset:.3f}"
        )

    @staticmethod
    def from_config(config: ConfigWrapper) -> KlipperScanModelConfiguration:
        name = config.get_name().split("scan_model", 1)[1].strip()
        coefficients = config.getfloatlist("coefficients")
        domain_raw = config.getfloatlist("domain", count=2)
        domain = (domain_raw[0], domain_raw[1])
        z_offset = config.getfloat("z_offset")

        return KlipperScanModelConfiguration(
            config,
            name=name,
            coefficients=coefficients,
            domain=domain,
            z_offset=z_offset,
        )


class KlipperTouchModelConfiguration(TouchModelConfiguration):
    def __init__(
        self,
        config: ConfigWrapper,
        *,
        name: str,
        threshold: int,
        speed: float,
        z_offset: float,
    ) -> None:
        self._config: ConfigWrapper = config
        self.name: str = name
        self.threshold: int = threshold
        self.speed: float = speed
        self.z_offset: float = z_offset

    @override
    def save_z_offset(self, new_offset: float) -> None:
        self._config.get_printer().lookup_object("configfile").set(
            self._config.get_name(), "z_offset", f"{new_offset:.3f}"
        )

    @staticmethod
    def from_config(config: ConfigWrapper) -> KlipperTouchModelConfiguration:
        name = config.get_name().split("touch_model", 1)[1].strip()
        threshold = config.getint("threshold", minval=1)
        speed = config.getfloat("speed", above=0)
        z_offset = config.getfloat("z_offset", maxval=0)
        return KlipperTouchModelConfiguration(
            config,
            name=name,
            threshold=threshold,
            speed=speed,
            z_offset=z_offset,
        )



from __future__ import annotations

from dataclasses import dataclass

from cartographer.probe.touch_mode import TouchBoundaries


@dataclass
class Configuration:
    mesh_min: tuple[float, float]
    mesh_max: tuple[float, float]
    x_offset: float
    y_offset: float

    move_speed: float = 10
    touch_samples: int = 3
    touch_max_samples: int = 10


def test_is_within_bounds():
    bounds = TouchBoundaries(min_x=10.0, max_x=20.0, min_y=5.0, max_y=15.0)

    # Test points clearly within bounds
    assert bounds.is_within(x=15.0, y=10.0) is True
    assert bounds.is_within(x=10.0, y=5.0) is True  # On lower bounds
    assert bounds.is_within(x=20.0, y=15.0) is True  # On upper bounds

    # Test points clearly outside bounds (beyond 0.01 tolerance)
    assert bounds.is_within(x=9.98, y=10.0) is False  # Left of min_x (outside epsilon)
    assert bounds.is_within(x=20.02, y=10.0) is False  # Right of max_x (outside epsilon)
    assert bounds.is_within(x=15.0, y=4.98) is False  # Below min_y (outside epsilon)
    assert bounds.is_within(x=15.0, y=15.02) is False  # Above max_y (outside epsilon)

    # Test points near boundary (within 0.01 tolerance)
    assert bounds.is_within(x=9.99, y=10.0) is True  # Just left of min_x (within epsilon)
    assert bounds.is_within(x=20.01, y=10.0) is True  # Just right of max_x (within epsilon)
    assert bounds.is_within(x=15.0, y=4.99) is True  # Just below min_y (within epsilon)
    assert bounds.is_within(x=15.0, y=15.01) is True  # Just above max_y (within epsilon)


def test_from_config():
    # Test with zero offsets
    config = Configuration(mesh_min=(0.0, 0.0), mesh_max=(100.0, 100.0), x_offset=0.0, y_offset=0.0)
    bounds = TouchBoundaries.from_config(config)
    assert bounds == TouchBoundaries(min_x=0.0, max_x=100.0, min_y=0.0, max_y=100.0)

    # Test with positive offsets
    config = Configuration(mesh_min=(0.0, 0.0), mesh_max=(100.0, 100.0), x_offset=10.0, y_offset=5.0)
    bounds = TouchBoundaries.from_config(config)
    assert bounds == TouchBoundaries(min_x=0.0, max_x=90.0, min_y=0.0, max_y=95.0)

    # Test with negative offsets
    config = Configuration(mesh_min=(0.0, 0.0), mesh_max=(100.0, 100.0), x_offset=-10.0, y_offset=-5.0)
    bounds = TouchBoundaries.from_config(config)
    assert bounds == TouchBoundaries(min_x=10.0, max_x=100.0, min_y=5.0, max_y=100.0)

    # Test with non-zero mesh_min
    config = Configuration(mesh_min=(10.0, 20.0), mesh_max=(100.0, 100.0), x_offset=5.0, y_offset=10.0)
    bounds = TouchBoundaries.from_config(config)
    assert bounds == TouchBoundaries(min_x=10.0, max_x=95.0, min_y=20.0, max_y=90.0)

    config = Configuration(mesh_min=(10.0, 10.0), mesh_max=(30.0, 30.0), x_offset=20, y_offset=20)
    bounds = TouchBoundaries.from_config(config)

    assert bounds.is_within(x=20.0, y=20.0) is False



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Callable, TypeVar

import pytest
from typing_extensions import ParamSpec, override

from cartographer.interfaces import TaskExecutor
from cartographer.printer_interface import MacroParams, Mcu, Position, Sample, Toolhead
from cartographer.stream import Session

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

P = ParamSpec("P")
R = TypeVar("R")


collect_ignore: list[str] = []
if sys.version_info < (3, 9):
    # pytest-bdd 8.0.0 requires Python 3.9+
    collect_ignore.append("bdd")


@pytest.fixture
def toolhead(mocker: MockerFixture) -> Toolhead:
    mock = mocker.MagicMock(spec=Toolhead, autospec=True, instance=True)

    def get_position() -> Position:
        return Position(x=10, y=10, z=5)

    def apply_axis_twist_compensation(position: Position) -> Position:
        return position

    mock.get_position = get_position
    mock.apply_axis_twist_compensation = apply_axis_twist_compensation

    return mock


@pytest.fixture
def session(mocker: MockerFixture) -> Session[Sample]:
    return Session(mocker.Mock(), mocker.Mock())


@pytest.fixture
def mcu(mocker: MockerFixture, session: Session[Sample]) -> Mcu[object, Sample]:
    mock = mocker.MagicMock(spec=Mcu, autospec=True, instance=True)
    mock.start_session = mocker.Mock(return_value=session)
    return mock


@pytest.fixture
def params(mocker: MockerFixture) -> MacroParams:
    return mocker.MagicMock(spec=MacroParams, autospec=True, instance=True)


class InlineTaskExecutor(TaskExecutor):
    @override
    def run(self, fn: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
        return fn(*args, **kwargs)


@pytest.fixture
def task_executor() -> TaskExecutor:
    return InlineTaskExecutor()



from cartographer.macros.probe import ProbeAccuracyMacro as ProbeAccuracyMacro
from cartographer.macros.probe import ProbeMacro as ProbeMacro
from cartographer.macros.probe import QueryProbeMacro as QueryProbeMacro
from cartographer.macros.probe import ZOffsetApplyProbeMacro as ZOffsetApplyProbeMacro



# https://github.com/Klipper3d/klipper/blob/master/klippy/extras/heaters.py
from typing import Callable, Protocol, TypedDict

class _Status(TypedDict):
    temperature: float
    target: float
    power: float

class Heater:
    def get_status(self, eventtime: float) -> _Status: ...

class _Sensor(Protocol):
    def setup_callback(self, temperature_callback: Callable[[float, float], None]) -> None: ...
    def get_report_time_delta(self) -> float: ...
    def setup_minmax(self, min_temp: float, max_temp: float) -> None: ...

class PrinterHeaters:
    def add_sensor_factory(self, sensor_type: str, sensor_factory: type[_Sensor]) -> None: ...



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Protocol


class ProbeMethod(Enum):
    SCAN = "scan"
    TOUCH = "touch"


@dataclass
class ScanModelFit:
    coefficients: list[float]
    domain: tuple[float, float]


class Configuration(Protocol):
    # [cartographer]
    x_offset: float
    y_offset: float
    backlash_compensation: float
    move_speed: float
    verbose: bool

    # [cartographer scan]
    scan_samples: int
    scan_mesh_runs: int

    # [cartographer touch]
    touch_samples: int
    touch_max_samples: int

    # [cartographer scan_model default]
    scan_models: dict[str, ScanModelConfiguration]

    # [cartographer touch_model default]
    touch_models: dict[str, TouchModelConfiguration]

    # [bed_mesh]
    mesh_min: tuple[float, float]
    mesh_max: tuple[float, float]
    scan_speed: float
    scan_height: float
    zero_reference_position: tuple[float, float]

    def save_new_scan_model(self, name: str, model: ScanModelFit) -> ScanModelConfiguration: ...
    def save_new_touch_model(self, name: str, speed: float, threshold: int) -> TouchModelConfiguration: ...


class ScanModelConfiguration(Protocol):
    name: str
    coefficients: list[float]
    domain: tuple[float, float]
    z_offset: float

    def save_z_offset(self, new_offset: float) -> None: ...


class TouchModelConfiguration(Protocol):
    name: str
    threshold: int
    speed: float
    z_offset: float

    def save_z_offset(self, new_offset: float) -> None: ...



# https://github.com/Klipper3d/klipper/blob/master/klippy/mcu.py

from typing import Any, Callable, Protocol, TypedDict, TypeVar, overload

from cffi import FFI
from klippy import Printer
from reactor import ReactorCompletion
from stepper import MCU_stepper

_T = TypeVar("_T")

class error(Exception): ...

class _MCUStatus(TypedDict):
    mcu_version: str

class _CommandQueue: ...

class MCU:
    error: type[error]

    def alloc_command_queue(self) -> _CommandQueue: ...
    def register_config_callback(self, callback: Callable[[], None]) -> None: ...
    def register_response(self, callback: Callable[[_T], None], message: str, oid: int | None = None) -> None: ...
    def lookup_command(self, msgformat: str, cq: _CommandQueue | None = None) -> CommandWrapper: ...
    def lookup_query_command(
        self,
        msgformat: str,
        respformat: str,
        oid: int | None = None,
        cq: _CommandQueue | None = None,
        is_async: bool = False,
    ) -> CommandQueryWrapper: ...
    def print_time_to_clock(self, print_time: float) -> int: ...
    def clock_to_print_time(self, clock: int) -> float: ...
    def clock32_to_clock64(self, clock32: int) -> int: ...
    def get_printer(self) -> Printer: ...
    def get_status(self) -> _MCUStatus: ...
    def is_fileoutput(self) -> bool: ...
    def get_constants(self) -> dict[str, float | int]: ...
    @overload
    def get_constant_float(self, name: str, default: float = ...) -> float: ...
    @overload
    def get_constant_float(self, name: str, default: None) -> float | None: ...

class MCU_trsync:
    REASON_ENDSTOP_HIT: int
    REASON_HOST_REQUEST: int
    REASON_PAST_END_TIME: int
    REASON_COMMS_TIMEOUT: int

    def __init__(self, mcu: MCU, trdispatch: FFI.CData) -> None: ...
    def get_oid(self) -> int: ...
    def get_mcu(self) -> MCU: ...
    def add_stepper(self, stepper: MCU_stepper) -> None: ...
    def get_steppers(self) -> list[MCU_stepper]: ...
    def start(
        self,
        print_time: float,
        report_offset: float,
        trigger_completion: ReactorCompletion,
        expire_timeout: float,
    ) -> None: ...
    def set_home_end_time(self, home_end_time: float) -> None: ...
    def stop(self) -> int: ...

class CommandWrapper:
    def send(self, data: list[int] = ..., minclock: int = 0, reqclock: int = 0) -> None: ...

class CommandQueryWrapper[T = Any]:
    def send(self, data: list[int] = ..., minclock: int = 0, reqclock: int = 0) -> T: ...

class TriggerDispatch:
    def __init__(self, mcu: MCU) -> None: ...
    def get_oid(self) -> int: ...
    def get_command_queue(self) -> _CommandQueue: ...
    def add_stepper(self, stepper: MCU_stepper) -> None: ...
    def get_steppers(self) -> list[MCU_stepper]: ...
    def start(self, print_time: float) -> ReactorCompletion: ...
    def wait_end(self, end_time: float) -> None: ...
    def stop(self) -> int: ...

class MCU_endstop(Protocol):
    def get_mcu(self) -> MCU: ...
    def add_stepper(self, stepper: MCU_stepper) -> None: ...
    def get_steppers(self) -> list[MCU_stepper]: ...
    def home_start(
        self,
        print_time: float,
        sample_time: float,
        sample_count: int,
        rest_time: float,
        triggered: bool = True,
    ) -> ReactorCompletion: ...
    def home_wait(self, home_end_time: float) -> float: ...
    def query_endstop(self, print_time: float) -> int: ...
    def get_position_endstop(self) -> float: ...

def get_printer_mcu(printer: Printer, name: str) -> MCU: ...



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, final

from extras.bed_mesh import BedMeshError
from gcode import GCodeCommand, GCodeDispatch
from typing_extensions import override

from cartographer.macros.bed_mesh import MeshHelper, MeshPoint

if TYPE_CHECKING:
    from configfile import ConfigWrapper

    from cartographer.printer_interface import Position

logger = logging.getLogger(__name__)


@final
class KlipperMeshHelper(MeshHelper[GCodeCommand]):
    def __init__(self, config: ConfigWrapper, gcode: GCodeDispatch) -> None:
        self.mesh_config = config.getsection("bed_mesh")
        self._bed_mesh = config.get_printer().load_object(self.mesh_config, "bed_mesh")
        # Loading "bed_mesh" above registers the command.
        self.macro = gcode.register_command("BED_MESH_CALIBRATE", None)

    @override
    def orig_macro(self, params: GCodeCommand) -> None:
        if self.macro is not None:
            self.macro(params)

    def _prepare(self, params: GCodeCommand) -> None:
        profile_name = params.get("PROFILE", "default")
        if not profile_name.strip():
            msg = "value for parameter 'PROFILE' must be specified"
            raise RuntimeError(msg)
        self._bed_mesh.set_mesh(None)
        self._bed_mesh.bmc._profile_name = profile_name
        try:
            self._bed_mesh.bmc.update_config(params)
        except BedMeshError as e:
            raise RuntimeError(str(e)) from e

    @override
    def prepare_scan_path(self, params: GCodeCommand) -> list[MeshPoint]:
        self._prepare(params)

        path = self._bed_mesh.bmc.probe_mgr.iter_rapid_path()
        return [MeshPoint(p[0], p[1], include) for (p, include) in path]

    @override
    def prepare_touch_points(
        self,
        params: GCodeCommand,
        *,
        mesh_min: tuple[float, float] | None,
        mesh_max: tuple[float, float] | None,
    ) -> list[MeshPoint]:
        raw = params.get_command_parameters()

        if mesh_min is not None and "MESH_MIN" not in raw:
            cfg_mesh_min = tuple([round(x, 2) for x in self.mesh_config.getfloatlist("mesh_min", count=2)])
            rounded_min = (round(mesh_min[0], 2), round(mesh_min[1], 2))
            if cfg_mesh_min != rounded_min:
                raw["MESH_MIN"] = f"{mesh_min[0]:.2f},{mesh_min[1]:.2f}"
                logger.info("Updating MESH_MIN from config value %s to %s", cfg_mesh_min, rounded_min)

        if mesh_max is not None and "MESH_MAX" not in raw:
            cfg_mesh_max = tuple(round(x, 2) for x in self.mesh_config.getfloatlist("mesh_max", count=2))
            rounded_max = (round(mesh_max[0], 2), round(mesh_max[1], 2))
            if cfg_mesh_max != rounded_max:
                raw["MESH_MAX"] = f"{mesh_max[0]:.2f},{mesh_max[1]:.2f}"
                logger.info("Updating MESH_MAX from config value %s to %s", cfg_mesh_max, rounded_max)

        self._prepare(params)

        points = self._bed_mesh.bmc.probe_mgr.get_std_path()
        if len(points) == 0:
            msg = "probe points are not set"
            raise RuntimeError(msg)
        return [MeshPoint(p[0], p[1], True) for p in points]

    @override
    def finalize(self, offset: Position, positions: list[Position]):
        self._bed_mesh.bmc.probe_finalize(
            [offset.x, offset.y, offset.z],
            [[p.x, p.y, p.z] for p in positions],
        )



import logging

logger = logging.getLogger(__name__)

formatter = logging.Formatter("[%(levelname)s:%(name)s] %(msg)s")
handler = logging.NullHandler()
handler.setFormatter(formatter)
logger.addHandler(handler)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

import pytest
from pytest_bdd import given, parsers, scenarios, then, when
from typing_extensions import TypeAlias

import cartographer.probe as probe
from cartographer.printer_interface import Position, Sample
from cartographer.probe import Probe

if TYPE_CHECKING:
    from unittest.mock import Mock

    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture

    from cartographer.printer_interface import MacroParams, Toolhead


scenarios("../../features/z_offset.feature")


ScanMode: TypeAlias = probe.ScanMode[object, Sample]
TouchMode: TypeAlias = probe.TouchMode[object]


@pytest.fixture
def scan(mocker: MockerFixture):
    mock = mocker.MagicMock(spec=probe.ScanMode, instance=True, autospec=True)
    mock.is_ready = True
    return mock


@pytest.fixture
def touch(mocker: MockerFixture):
    mock = mocker.MagicMock(spec=probe.TouchMode, instance=True, autospec=True)
    mock.is_ready = False
    return mock


@given("a probe", target_fixture="probe")
def given_probe(scan: ScanMode, touch: TouchMode) -> Probe:
    return Probe(scan, touch)


@given("the probe has touch ready")
def given_probe_touch_model_loaded(touch: Mock):
    touch.is_ready = True


@given(parsers.parse("the probe's current z-offset is {offset:g}"))
def given_current_z_offset(scan: Mock, touch: Mock, offset: float):
    scan.offset = Position(0, 0, offset)
    touch.offset = Position(0, 0, offset)


@given(parsers.parse("I have baby stepped the nozzle {offset:g}mm up"))
def given_baby_step_up(toolhead: Toolhead, offset: float):
    toolhead.get_gcode_z_offset = lambda: offset


@given(parsers.parse("I have baby stepped the nozzle {offset:g}mm down"))
def given_baby_step_down(toolhead: Toolhead, offset: float):
    toolhead.get_gcode_z_offset = lambda: -offset


@when("I run the Z_OFFSET_APPLY_PROBE macro")
def when_run_probe_accuracy_macro(params: MacroParams, caplog: LogCaptureFixture, probe: Probe, toolhead: Toolhead):
    from cartographer.macros.probe import ZOffsetApplyProbeMacro

    macro = ZOffsetApplyProbeMacro(probe, toolhead)
    with caplog.at_level(logging.INFO):
        macro.run(params)


@then("it should set z-offset on the scan model")
def then_set_scan_z_offset(scan: ScanMode):
    assert cast("Mock", scan.save_z_offset).call_count == 1


@then("it should not set z-offset on the scan model")
def then_not_set_scan_z_offset(scan: ScanMode):
    assert cast("Mock", scan.save_z_offset).call_count == 0


@then("it should set z-offset on the touch model")
def then_set_touch_z_offset(touch: TouchMode):
    assert cast("Mock", touch.save_z_offset).call_count == 1


@then(parsers.parse("it should set scan z-offset to {offset:g}"))
def then_update_scan_z_offset(mocker: MockerFixture, scan: ScanMode, offset: str):
    assert cast("Mock", scan.save_z_offset).mock_calls == [mocker.call(offset)]


@then(parsers.parse("it should set touch z-offset to {offset:g}"))
def then_update_touch_z_offset(mocker: MockerFixture, touch: TouchMode, offset: str):
    assert cast("Mock", touch.save_z_offset).mock_calls == [mocker.call(offset)]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, final

from typing_extensions import override

from cartographer.macros.axis_twist_compensation import AxisTwistCompensationHelper, CalibrationOptions

if TYPE_CHECKING:
    from configfile import ConfigWrapper


@final
class KlipperAxisTwistCompensationHelper(AxisTwistCompensationHelper):
    def __init__(self, config: ConfigWrapper) -> None:
        self.config = config.getsection("axis_twist_compensation")
        self.compensation = config.get_printer().load_object(self.config, "axis_twist_compensation")
        self.move_height = self.compensation.horizontal_move_z
        self.speed = self.compensation.speed

    @override
    def clear_compensations(self, axis: Literal["x", "y"]) -> None:
        self.compensation.clear_compensations(axis.upper())

    @override
    def save_compensations(self, axis: Literal["x", "y"], start: float, end: float, values: list[float]) -> None:
        configfile = self.config.get_printer().lookup_object("configfile")
        configname = self.config.get_name()
        values_as_str = ", ".join([f"{x:.6f}" for x in values])

        if axis == "x":
            configfile.set(configname, "z_compensations", values_as_str)
            configfile.set(configname, "compensation_start_x", start)
            configfile.set(configname, "compensation_end_x", end)

            self.compensation.z_compensations = values
            self.compensation.compensation_start_x = start
            self.compensation.compensation_end_x = end
        elif axis == "y":
            configfile.set(configname, "zy_compensations", values_as_str)
            configfile.set(configname, "compensation_start_y", start)
            configfile.set(configname, "compensation_end_y", end)

            self.compensation.zy_compensations = values
            self.compensation.compensation_start_y = start
            self.compensation.compensation_end_y = end

    @override
    def get_calibration_options(self, axis: Literal["x", "y"]) -> CalibrationOptions:
        if axis == "x":
            return CalibrationOptions(
                self.compensation.calibrate_start_x,
                self.compensation.calibrate_end_x,
                self.compensation.calibrate_y,
            )
        elif axis == "y":
            return CalibrationOptions(
                self.compensation.calibrate_start_y,
                self.compensation.calibrate_end_y,
                self.compensation.calibrate_x,
            )



from cartographer.probe.probe import Probe as Probe
from cartographer.probe.scan_mode import ScanMode as ScanMode
from cartographer.probe.scan_model import ScanModel as ScanModel
from cartographer.probe.touch_mode import TouchMode as TouchMode



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict, final

import pins

if TYPE_CHECKING:
    from klippy import Printer
    from mcu import MCU_endstop


# https://github.com/Klipper3d/klipper/blob/75a10bfcafc0655e36e5ecf01c3f2033a14ef2c7/klippy/pins.py#L93-L94
class _PinParams(TypedDict):
    pin: str
    invert: int
    pullup: int


@final
class CartographerHomingChip:
    def __init__(self, printer: Printer, endstop: MCU_endstop) -> None:
        self.endstop = endstop

    def setup_pin(self, pin_type: str, pin_params: _PinParams) -> MCU_endstop:
        if pin_type != "endstop" or pin_params["pin"] != "z_virtual_endstop":
            msg = "cartographer virtual endstop only useful as endstop pin"
            raise pins.error(msg)
        if pin_params["invert"] or pin_params["pullup"]:
            msg = "can not pullup/invert probe virtual endstop"
            raise pins.error(msg)

        return self.endstop



from __future__ import annotations

import logging
import math
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, Protocol, final

import numpy as np
from typing_extensions import override

from cartographer.lib.nearest_neighbor import NearestNeighborSearcher
from cartographer.printer_interface import C, Macro, MacroParams, P, Position, S, Toolhead

if TYPE_CHECKING:
    from cartographer.interfaces import TaskExecutor
    from cartographer.probe import Probe
    from cartographer.probe.scan_mode import Model, ScanMode
    from cartographer.probe.touch_mode import TouchMode

logger = logging.getLogger(__name__)


class Configuration(Protocol):
    scan_speed: float
    scan_height: float
    scan_mesh_runs: int


@dataclass
class MeshPoint:
    x: float
    y: float
    include: bool


class MeshHelper(Generic[P], Protocol):
    def orig_macro(self, params: P) -> None: ...
    def prepare_scan_path(self, params: P) -> list[MeshPoint]: ...
    def prepare_touch_points(
        self,
        params: P,
        *,
        mesh_min: tuple[float, float] | None,
        mesh_max: tuple[float, float] | None,
    ) -> list[MeshPoint]: ...
    def finalize(self, offset: Position, positions: list[Position]): ...


@final
class BedMeshCalibrateMacro(Macro[P]):
    name = "BED_MESH_CALIBRATE"
    description = "Gather samples across the bed to calibrate the bed mesh."

    def __init__(
        self,
        probe: Probe,
        toolhead: Toolhead,
        helper: MeshHelper[P],
        task_executor: TaskExecutor,
        config: Configuration,
    ) -> None:
        self.helper = helper
        self.probe = probe
        self.scan_mesh = _ScanMeshRunner(probe.scan, toolhead, task_executor, config)
        self.touch_mesh = _TouchMeshRunner(probe.touch, toolhead, config)

    @override
    def run(self, params: P) -> None:
        method = params.get("METHOD", default="scan").lower()
        if method != "scan" and method != "rapid_scan" and method != "touch":
            return self.helper.orig_macro(params)

        start_time = time.time()
        if method == "touch":
            boundaries = self.probe.touch.boundaries
            offset, positions = self.touch_mesh.run(
                params,
                self.helper.prepare_touch_points(
                    params,
                    mesh_min=(boundaries.min_x, boundaries.min_y),
                    mesh_max=(boundaries.max_x, boundaries.max_y),
                ),
            )
        else:
            offset, positions = self.scan_mesh.run(params, self.helper.prepare_scan_path(params))
        logger.debug("Bed mesh completed in %.2f seconds", time.time() - start_time)

        self.helper.finalize(offset, positions)


@final
class _TouchMeshRunner:
    def __init__(
        self,
        probe: TouchMode[object],
        toolhead: Toolhead,
        config: Configuration,
    ) -> None:
        self.probe = probe
        self.toolhead = toolhead
        self.config = config

    def run(self, params: MacroParams, points: list[MeshPoint]) -> tuple[Position, list[Position]]:
        speed = params.get_float("SPEED", default=self.config.scan_speed, minval=1)
        move_height = params.get_float("HORIZONTAL_MOVE_Z", default=self.config.scan_height, minval=1)
        if self.probe.model is None:
            msg = "cannot run bed mesh without a model"
            raise RuntimeError(msg)
        points = [p for p in points if p.include]

        for p in points:
            if not self.probe.is_within_boundaries(x=p.x, y=p.y):
                msg = f"probe point ({p.x:.2f},{p.y:.2f}) is outside of the touch boundaries "
                raise RuntimeError(msg)

        self.toolhead.move(z=move_height, speed=5)
        positions: list[Position] = []
        for p in points:
            self.toolhead.move(x=p.x, y=p.y, speed=speed)
            trigger_pos = self.probe.perform_probe()
            positions.append(Position(p.x, p.y, trigger_pos))

        return self.probe.offset, positions


@final
class _ScanMeshRunner:
    def __init__(
        self,
        probe: ScanMode[C, S],
        toolhead: Toolhead,
        task_executor: TaskExecutor,
        config: Configuration,
    ) -> None:
        self.probe = probe
        self.toolhead = toolhead
        self.task_executor = task_executor
        self.config = config

    def run(self, params: MacroParams, path: list[MeshPoint]) -> tuple[Position, list[Position]]:
        runs = params.get_int("RUNS", default=self.config.scan_mesh_runs, minval=1)
        speed = params.get_float("SPEED", default=self.config.scan_speed, minval=1)
        scan_height = params.get_float("HORIZONTAL_MOVE_Z", default=self.config.scan_height, minval=1)
        if self.probe.model is None:
            msg = "cannot run bed mesh without a model"
            raise RuntimeError(msg)

        self.toolhead.move(z=scan_height, speed=5)
        self._move_to_point(path[0], speed)

        start_time = time.time()
        with self.probe.start_session() as session:
            session.wait_for(lambda samples: len(samples) >= 5)
            for i in range(runs):
                is_odd = i & 1  # Bitwise check for odd numbers
                # Every other run should be going in reverse
                path_iter = reversed(path) if is_odd else path
                for point in path_iter:
                    self._move_to_point(point, speed)
                self.toolhead.dwell(0.250)
                self.toolhead.wait_moves()
            move_time = self.toolhead.get_last_move_time()
            session.wait_for(lambda samples: samples[-1].time >= move_time)
            count = len(session.items)
            session.wait_for(lambda samples: len(samples) >= count + 50)
        logger.debug("Bed scan completed in %.2f seconds", time.time() - start_time)

        samples = session.get_items()
        logger.debug("Gathered %d samples", len(samples))

        positions = self.task_executor.run(self._calculate_positions, self.probe.model, path, samples, scan_height)
        return self.probe.offset, positions

    def _move_to_point(self, point: MeshPoint, speed: float) -> None:
        offset = self.probe.offset
        self.toolhead.move(x=point.x - offset.x, y=point.y - offset.y, speed=speed)

    def _key(self, point: MeshPoint) -> tuple[float, float]:
        return round(point.x, 2), round(point.y, 2)

    def _calculate_positions(
        self, model: Model, path: list[MeshPoint], samples: list[S], scan_height: float
    ) -> list[Position]:
        included_points = [p for p in path if p.include]

        start_time = time.time()
        clusters = self._build_clusters(
            samples,
            included_points,
        )
        logger.debug("Sample clustering completed in %.2f seconds", time.time() - start_time)

        start_time = time.time()
        positions = [
            self._compute_position(
                (x, y),
                cluster,
                model,
                scan_height,
            )
            for (x, y), cluster in clusters.items()
        ]
        logger.debug("Cluster position computation completed in %.2f seconds", time.time() - start_time)
        return positions

    def _build_clusters(
        self,
        samples: list[S],
        points: list[MeshPoint],
    ) -> dict[tuple[float, float], list[S]]:
        offset = self.probe.offset
        searcher = NearestNeighborSearcher(points)
        cluster_map: dict[tuple[float, float], list[S]] = {self._key(p): [] for p in points}

        valid_samples: list[S] = []
        adjusted_positions: list[tuple[float, float]] = []

        for s in samples:
            if s.position is None:
                continue
            valid_samples.append(s)
            adjusted_positions.append(
                (
                    s.position.x + offset.x,
                    s.position.y + offset.y,
                )
            )

        nearest_points = searcher.batch_query(adjusted_positions)

        for s, point in zip(valid_samples, nearest_points):
            if point is not None and point.include:
                key = self._key(point)
                cluster_map[key].append(s)

        return cluster_map

    def _compute_position(
        self,
        key: tuple[float, float],
        cluster: list[S],
        model: Model,
        scan_height: float,
    ) -> Position:
        offset = self.probe.offset
        x, y = key
        if not cluster:
            msg = f"cluster ({x:.2f},{y:.2f}) has no samples"
            raise RuntimeError(msg)

        distances = list(map(lambda s: model.frequency_to_distance(s.frequency), cluster))
        median_distance = float(np.median(distances))

        if not math.isfinite(median_distance):
            msg = f"cluster ({x:.2f},{y:.2f}) has no valid samples"
            raise RuntimeError(msg)

        trigger_z = scan_height + self.probe.probe_height - median_distance
        pos = Position(x - offset.x, y - offset.y, trigger_z)
        return self.toolhead.apply_axis_twist_compensation(pos)



from __future__ import annotations

import logging
import math
from typing import TYPE_CHECKING, Generic, Protocol

import numpy as np
from typing_extensions import override

from cartographer.printer_interface import C, Endstop, HomingState, Position, ProbeMode, S

if TYPE_CHECKING:
    from cartographer.printer_interface import Mcu, Toolhead
    from cartographer.stream import Session

logger = logging.getLogger(__name__)


class Model(Protocol):
    @property
    def name(self) -> str: ...

    @property
    def z_offset(self) -> float: ...
    def save_z_offset(self, new_offset: float) -> None: ...
    def distance_to_frequency(self, distance: float) -> float: ...
    def frequency_to_distance(self, frequency: float) -> float: ...


TRIGGER_DISTANCE = 2.0


class Configuration(Protocol):
    x_offset: float
    y_offset: float
    move_speed: float

    scan_samples: int


class ScanMode(ProbeMode, Endstop[C], Generic[C, S]):
    """Implementation for Scan mode."""

    def get_model(self) -> Model:
        if self.model is None:
            msg = "no scan model loaded"
            raise RuntimeError(msg)
        return self.model

    @property
    @override
    def offset(self) -> Position:
        z_offset = self.model.z_offset if self.model else 0.0
        return Position(self.config.x_offset, self.config.y_offset, self.probe_height + z_offset)

    @override
    def save_z_offset(self, new_offset: float) -> None:
        self.get_model().save_z_offset(new_offset - self.probe_height)

    @property
    @override
    def is_ready(self) -> bool:
        return self.model is not None

    def __init__(
        self,
        mcu: Mcu[C, S],
        toolhead: Toolhead,
        config: Configuration,
        *,
        model: Model | None = None,
        probe_height: float = TRIGGER_DISTANCE,
    ) -> None:
        self._toolhead: Toolhead = toolhead
        self.model: Model | None = model
        self.config: Configuration = config
        self.probe_height: float = probe_height
        self._mcu: Mcu[C, S] = mcu

    @override
    def perform_probe(self) -> float:
        if not self._toolhead.is_homed("z"):
            msg = "z axis must be homed before probing"
            raise RuntimeError(msg)
        self._toolhead.move(z=self.probe_height, speed=self.config.move_speed)
        self._toolhead.wait_moves()
        toolhead_pos = self._toolhead.get_position()
        dist = toolhead_pos.z + self.probe_height - self.measure_distance()

        if math.isinf(dist):
            msg = "toolhead stopped outside model range"
            raise RuntimeError(msg)

        pos = self._toolhead.apply_axis_twist_compensation(Position(toolhead_pos.x, toolhead_pos.y, dist))
        logger.info("probe at %.3f,%.3f is z=%.6f", pos.x, pos.y, pos.z)
        return pos.z

    def distance_to_frequency(self, distance: float) -> float:
        if self.model is None:
            msg = "cannot convert distance to frequency without a model"
            raise RuntimeError(msg)
        return self.model.distance_to_frequency(distance)

    def measure_distance(
        self, *, time: float | None = None, min_sample_count: int | None = None, skip_count: int = 5
    ) -> float:
        min_sample_count = min_sample_count or self.config.scan_samples
        if self.model is None:
            msg = "cannot measure distance without a model"
            raise RuntimeError(msg)
        time = time or self._toolhead.get_last_move_time()

        with self._mcu.start_session(lambda sample: sample.time >= time) as session:
            session.wait_for(lambda samples: len(samples) >= min_sample_count + skip_count)
        samples = session.get_items()[skip_count:]

        dist = float(np.median([self.model.frequency_to_distance(sample.frequency) for sample in samples]))
        return dist

    @override
    def query_is_triggered(self, print_time: float) -> bool:
        distance = self.measure_distance(time=print_time)
        return distance <= self.get_endstop_position()

    @override
    def get_endstop_position(self) -> float:
        return self.probe_height

    @override
    def home_start(self, print_time: float) -> C:
        trigger_frequency = self.distance_to_frequency(self.get_endstop_position())
        return self._mcu.start_homing_scan(print_time, trigger_frequency)

    @override
    def on_home_end(self, homing_state: HomingState) -> None:
        if self not in homing_state.endstops:
            return
        if not homing_state.is_homing_z():
            return
        distance = self.measure_distance()
        if math.isinf(distance):
            msg = "toolhead stopped outside model range"
            raise RuntimeError(msg)

        homing_state.set_z_homed_position(distance)

    @override
    def home_wait(self, home_end_time: float) -> float:
        return self._mcu.stop_homing(home_end_time)

    def start_session(
        self,
    ) -> Session[S]:
        time = self._toolhead.get_last_move_time()
        return self._mcu.start_session(lambda sample: sample.time >= time)



# https://github.com/Klipper3d/klipper/blob/master/klippy/klippy.py
from typing import Callable, Literal, overload

import configfile
from configfile import ConfigWrapper, PrinterConfig
from gcode import CommandError, GCodeDispatch
from gcode_move import GCodeMove
from pins import PrinterPins
from reactor import Reactor
from stepper import PrinterRail
from toolhead import ToolHead

from cartographer.klipper.extra import PrinterCartographer
from extras.axis_twist_compensation import AxisTwistCompensation
from extras.bed_mesh import BedMesh
from extras.heaters import PrinterHeaters
from extras.homing import Homing, HomingMove, PrinterHoming
from extras.motion_report import PrinterMotionReport
from extras.probe import PrinterProbe

class Printer:
    config_error: type[configfile.error]
    command_error: type[CommandError]
    def add_object(self, name: Literal["probe"], obj: PrinterProbe) -> None: ...
    @overload
    def load_object(
        self,
        config: ConfigWrapper,
        section: Literal["bed_mesh"],
    ) -> BedMesh: ...
    @overload
    def load_object(
        self,
        config: ConfigWrapper,
        section: Literal["heaters"],
    ) -> PrinterHeaters: ...
    @overload
    def load_object(
        self,
        config: ConfigWrapper,
        section: Literal["motion_report"],
    ) -> PrinterMotionReport: ...
    @overload
    def load_object(
        self,
        config: ConfigWrapper,
        section: Literal["axis_twist_compensation"],
    ) -> AxisTwistCompensation: ...
    def is_shutdown(self) -> bool: ...
    def invoke_shutdown(self, msg: str) -> None: ...
    def get_reactor(self) -> Reactor: ...
    @overload
    def register_event_handler(self, event: Literal["klippy:connect"], callback: Callable[[], None]) -> None: ...
    @overload
    def register_event_handler(self, event: Literal["klippy:disconnect"], callback: Callable[[], None]) -> None: ...
    @overload
    def register_event_handler(self, event: Literal["klippy:shutdown"], callback: Callable[[], None]) -> None: ...
    @overload
    def register_event_handler(self, event: Literal["klippy:mcu_identify"], callback: Callable[[], None]) -> None: ...
    @overload
    def register_event_handler(
        self,
        event: Literal["homing:home_rails_begin"],
        callback: Callable[[Homing, list[PrinterRail]], None],
    ) -> None: ...
    @overload
    def register_event_handler(
        self,
        event: Literal["homing:home_rails_end"],
        callback: Callable[[Homing, list[PrinterRail]], None],
    ) -> None: ...
    @overload
    def register_event_handler(
        self,
        event: Literal["homing:homing_move_begin"],
        callback: Callable[[HomingMove], None],
    ) -> None: ...
    @overload
    def register_event_handler(
        self,
        event: Literal["homing:homing_move_end"],
        callback: Callable[[HomingMove], None],
    ) -> None: ...
    @overload
    def lookup_object(self, name: Literal["bed_mesh"]) -> BedMesh: ...
    @overload
    def lookup_object(self, name: Literal["configfile"]) -> PrinterConfig: ...
    @overload
    def lookup_object(self, name: Literal["gcode"]) -> GCodeDispatch: ...
    @overload
    def lookup_object(self, name: Literal["gcode_move"]) -> GCodeMove: ...
    @overload
    def lookup_object(self, name: Literal["homing"]) -> PrinterHoming: ...
    @overload
    def lookup_object(self, name: Literal["motion_report"]) -> PrinterMotionReport: ...
    @overload
    def lookup_object(self, name: Literal["pins"]) -> PrinterPins: ...
    @overload
    def lookup_object(self, name: Literal["toolhead"]) -> ToolHead: ...
    @overload
    def lookup_object(self, name: Literal["cartographer"]) -> PrinterCartographer: ...
    def send_event(self, event: Literal["probe:update_results"], pos: list[float]) -> None: ...



import logging
from typing import Callable

import pytest
from typing_extensions import TypeAlias

from cartographer.klipper.logging import GCodeConsoleFilter, GCodeConsoleFormatter, format_macro


@pytest.fixture
def formatter() -> GCodeConsoleFormatter:
    return GCodeConsoleFormatter()


LogRecordFactory: TypeAlias = Callable[[int, str], logging.LogRecord]


@pytest.fixture
def log_record() -> LogRecordFactory:
    def _make(level: int, msg: str, name: str = "test.logger"):
        return logging.LogRecord(name=name, level=level, pathname="test", lineno=1, msg=msg, args=(), exc_info=None)

    return _make


@pytest.mark.parametrize(
    "message,expected",
    [
        ("TOUCH", format_macro("TOUCH")),
        ("TOUCH_CALIBRATE THRESHOLD_START=1000", format_macro("TOUCH_CALIBRATE THRESHOLD_START=1000")),
        ("Start TOUCH then SAVE_CONFIG", f"Start {format_macro('TOUCH')} then {format_macro('SAVE_CONFIG')}"),
        ("Normal message", "Normal message"),
        (
            "Multiple TOUCH X=1 Y=2 SAVE_CONFIG",
            f"Multiple {format_macro('TOUCH X=1 Y=2')} {format_macro('SAVE_CONFIG')}",
        ),
    ],
)
def test_macro_formatting_info(
    formatter: GCodeConsoleFormatter, log_record: LogRecordFactory, message: str, expected: str
):
    record = log_record(logging.INFO, message)
    output = formatter.format(record)
    assert output == expected


@pytest.mark.parametrize("level", [logging.DEBUG, logging.INFO, logging.WARNING])
def test_formatting_no_prefix(formatter: GCodeConsoleFormatter, level: int, log_record: LogRecordFactory):
    record = log_record(level, "test message")
    output = formatter.format(record)
    assert output == "test message"


def test_formatting_error_prefix(formatter: GCodeConsoleFormatter, log_record: LogRecordFactory):
    record = log_record(logging.ERROR, "test error")
    output = formatter.format(record)
    assert output == "!! test error"


@pytest.mark.parametrize(
    "logger_name,level,expected",
    [
        ("klipper.mcu", logging.DEBUG, False),  # filtered out
        ("klipper.mcu", logging.INFO, False),  # filtered out
        ("klipper.mcu", logging.WARNING, True),  # allowed
        ("klipper.mcu", logging.ERROR, True),  # allowed
        ("klipper.printer", logging.INFO, True),  # allowed
        ("macros.scan", logging.DEBUG, True),  # allowed
    ],
)
def test_console_filter_behavior(logger_name: str, level: int, expected: bool):
    record = logging.LogRecord(
        name=logger_name, level=level, pathname="test", lineno=1, msg="test", args=(), exc_info=None
    )
    filt = GCodeConsoleFilter()
    assert filt.filter(record) is expected



from __future__ import annotations

import logging
from enum import IntEnum
from typing import TYPE_CHECKING, NamedTuple, final

if TYPE_CHECKING:
    from mcu import MCU, CommandWrapper

logger = logging.getLogger(__name__)


class TriggerMethod(IntEnum):
    SCAN = 0
    TOUCH = 1


class HomeCommand(NamedTuple):
    trsync_oid: int
    trigger_reason: int
    trigger_invert: int
    threshold: int
    trigger_method: TriggerMethod


class ThresholdCommand(NamedTuple):
    trigger: int
    untrigger: int


@final
class KlipperCartographerCommands:
    def __init__(self, mcu: MCU):
        self._command_queue = mcu.alloc_command_queue()
        self._stream_command = mcu.lookup_command("cartographer_stream en=%u", cq=self._command_queue)
        self._set_threshold_command = mcu.lookup_command(
            "cartographer_set_threshold trigger=%u untrigger=%u", cq=self._command_queue
        )
        self._start_home_command = mcu.lookup_command(
            "cartographer_home trsync_oid=%c trigger_reason=%c trigger_invert=%c threshold=%u trigger_method=%u",
            cq=self._command_queue,
        )
        self._stop_home_command = mcu.lookup_command("cartographer_stop_home", cq=self._command_queue)

    def _ensure_initialized(self, command: CommandWrapper | None, name: str) -> CommandWrapper:
        if command is None:
            msg = f"command {name} has not been initialized"
            raise RuntimeError(msg)
        return command

    def send_stream_state(self, *, enable: bool) -> None:
        command = self._ensure_initialized(self._stream_command, "stream command")
        logger.debug("%s stream", "Starting" if enable else "Stopping")
        command.send([1 if enable else 0])

    def send_threshold(self, command: ThresholdCommand) -> None:
        cmd = self._ensure_initialized(self._set_threshold_command, "set threshold command")
        logger.debug("Sending trigger frequency threshold command %s", list(command))
        cmd.send(list(command))

    def send_home(self, command: HomeCommand) -> None:
        cmd = self._ensure_initialized(self._start_home_command, "start home command")
        logger.debug("Sending home command %s", list(command))
        cmd.send(list(command))

    def send_stop_home(self) -> None:
        cmd = self._ensure_initialized(self._stop_home_command, "stop home command")
        logger.debug("Sending stop home command")
        cmd.send()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, final

from typing_extensions import override

from cartographer.klipper.utils import reraise_as_command_error
from cartographer.printer_interface import HomingAxis, HomingState

if TYPE_CHECKING:
    from collections.abc import Sequence

    from extras.homing import Homing
    from mcu import MCU
    from reactor import ReactorCompletion
    from stepper import MCU_stepper, PrinterRail

    from cartographer.klipper.mcu import KlipperCartographerMcu
    from cartographer.printer_interface import Endstop

logger = logging.getLogger(__name__)

axis_mapping: dict[HomingAxis, int] = {
    "x": 0,
    "y": 1,
    "z": 2,
}


def axis_to_index(axis: HomingAxis) -> int:
    return axis_mapping[axis]


@final
class KlipperHomingState(HomingState):
    def __init__(self, homing: Homing, endstops: Sequence[Endstop[object]]) -> None:
        self.homing = homing
        self.endstops = endstops

    @override
    def is_homing_z(self) -> bool:
        return axis_to_index("z") in self.homing.get_axes()

    @override
    def set_z_homed_position(self, position: float) -> None:
        logger.debug("Setting homed distance for z to %.3f", position)
        self.homing.set_homed_position([None, None, position])


class _MemoizedEndstop(type):
    _endstops: dict[Endstop[ReactorCompletion], KlipperEndstop] = {}

    @override
    def __call__(cls, mcu: KlipperCartographerMcu, endstop: Endstop[ReactorCompletion]):
        if endstop not in cls._endstops:
            cls._endstops[endstop] = super().__call__(mcu, endstop)
        return cls._endstops[endstop]


@final
class KlipperEndstop(metaclass=_MemoizedEndstop):
    def __init__(self, mcu: KlipperCartographerMcu, endstop: Endstop[ReactorCompletion]):
        self.printer = mcu.klipper_mcu.get_printer()
        self.mcu = mcu
        self.endstop = endstop
        self.printer.register_event_handler("homing:home_rails_end", self.home_rails_end)

    @reraise_as_command_error
    def home_rails_end(self, homing: Homing, rails: list[PrinterRail]) -> None:
        endstops = [es.endstop for rail in rails for es, _ in rail.get_endstops() if isinstance(es, KlipperEndstop)]
        self.endstop.on_home_end(KlipperHomingState(homing, endstops))

    def get_mcu(self) -> MCU:
        return self.mcu.klipper_mcu

    def add_stepper(self, stepper: MCU_stepper) -> None:
        return self.mcu.dispatch.add_stepper(stepper)

    def get_steppers(self) -> list[MCU_stepper]:
        return self.mcu.dispatch.get_steppers()

    @reraise_as_command_error
    def home_start(
        self,
        print_time: float,
        sample_time: float,
        sample_count: int,
        rest_time: float,
        triggered: bool = True,
    ) -> ReactorCompletion:
        del sample_time, sample_count, rest_time, triggered
        return self.endstop.home_start(print_time)

    @reraise_as_command_error
    def home_wait(self, home_end_time: float) -> float:
        return self.endstop.home_wait(home_end_time)

    @reraise_as_command_error
    def query_endstop(self, print_time: float) -> int:
        return 1 if self.endstop.query_is_triggered(print_time) else 0

    def get_position_endstop(self) -> float:
        return self.endstop.get_endstop_position()



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from cartographer.printer_interface import MacroParams

T = TypeVar("T", bound=Enum)


def get_enum_choice(params: MacroParams, option: str, enum_type: type[T], default: T) -> T:
    choice = params.get(option, default=default.value)

    # Convert both the choice and enum values to lowercase for case-insensitive comparison
    lower_choice = str(choice).lower()
    lower_mapping = {str(v.value).lower(): v for v in enum_type}

    if lower_choice not in lower_mapping:
        msg = f"invalid choice '{choice}' for option '{option}'"
        raise RuntimeError(msg)

    return lower_mapping[lower_choice]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, final

import numpy as np
from typing_extensions import override

from cartographer.printer_interface import Macro, MacroParams

if TYPE_CHECKING:
    from cartographer.printer_interface import Toolhead
    from cartographer.probe import Probe

logger = logging.getLogger(__name__)


@final
class ProbeMacro(Macro[MacroParams]):
    name = "PROBE"
    description = "Probe the bed to get the height offset at the current position."
    last_trigger_position: float | None = None

    def __init__(self, probe: Probe) -> None:
        self._probe = probe

    @override
    def run(self, params: MacroParams) -> None:
        trigger_pos = self._probe.perform_scan()
        logger.info("Result is z=%.6f", trigger_pos)
        self.last_trigger_position = trigger_pos


@final
class ProbeAccuracyMacro(Macro[MacroParams]):
    name = "PROBE_ACCURACY"
    description = "Probe the bed multiple times to measure the accuracy of the probe."

    def __init__(self, probe: Probe, toolhead: Toolhead) -> None:
        self._probe = probe
        self._toolhead = toolhead

    @override
    def run(self, params: MacroParams) -> None:
        lift_speed = params.get_float("LIFT_SPEED", 5.0, above=0)
        retract = params.get_float("SAMPLE_RETRACT_DIST", 1.0, minval=0)
        sample_count = params.get_int("SAMPLES", 10, minval=1)
        position = self._toolhead.get_position()

        logger.info(
            "PROBE_ACCURACY at X:%.3f Y:%.3f Z:%.3f (samples=%d retract=%.3f lift_speed=%.1f)",
            position.x,
            position.y,
            position.z,
            sample_count,
            retract,
            lift_speed,
        )

        self._toolhead.move(z=position.z + retract, speed=lift_speed)
        measurements: list[float] = []
        while len(measurements) < sample_count:
            trigger_pos = self._probe.perform_scan()
            measurements.append(trigger_pos)
            pos = self._toolhead.get_position()
            self._toolhead.move(z=pos.z + retract, speed=lift_speed)
        logger.debug("Measurements gathered: %s", measurements)

        max_value = max(measurements)
        min_value = min(measurements)
        range_value = max_value - min_value
        avg_value = np.mean(measurements)
        median = np.median(measurements)
        std_dev = np.std(measurements)

        logger.info(
            """
            probe accuracy results: maximum %.6f, minimum %.6f, range %.6f,
            average %.6f, median %.6f, standard deviation %.6f
            """,
            max_value,
            min_value,
            range_value,
            avg_value,
            median,
            std_dev,
        )


@final
class QueryProbeMacro(Macro[MacroParams]):
    name = "QUERY_PROBE"
    description = "Return the status of the z-probe"
    last_triggered: bool = False

    def __init__(self, probe: Probe) -> None:
        self._probe = probe

    @override
    def run(self, params: MacroParams) -> None:
        triggered = self._probe.query_is_triggered()
        logger.info("probe: %s", "TRIGGERED" if triggered else "open")
        self.last_triggered = triggered


@final
class ZOffsetApplyProbeMacro(Macro[MacroParams]):
    name = "Z_OFFSET_APPLY_PROBE"
    description = "Adjust the probe's z_offset"

    def __init__(self, probe: Probe, toolhead: Toolhead) -> None:
        self._probe = probe
        self._toolhead = toolhead

    @override
    def run(self, params: MacroParams) -> None:
        additional_offset = self._toolhead.get_gcode_z_offset()
        # If a touch is ready, we assume the user uses touch.
        probe_mode_str, probe_mode = (
            ("touch", self._probe.touch) if self._probe.touch.is_ready else ("scan", self._probe.scan)
        )
        current_offset = probe_mode.offset.z
        new_offset = current_offset - additional_offset
        logger.info(
            """cartographer: %s %s z_offset: %.3f
            The SAVE_CONFIG command will update the printer config file
            with the above and restart the printer.""",
            probe_mode_str,
            probe_mode.get_model().name,
            new_offset,
        )

        probe_mode.save_z_offset(new_offset)



from __future__ import annotations

import threading
import time
from typing import Callable

import pytest
from typing_extensions import override

from cartographer.stream import Condition, Stream


class MockCondition(Condition):
    def __init__(self):
        self._condition: threading.Condition = threading.Condition()

    @override
    def notify_all(self) -> None:
        with self._condition:
            self._condition.notify_all()

    @override
    def wait_for(self, predicate: Callable[[], bool]) -> None:
        with self._condition:
            _ = self._condition.wait_for(predicate)


class MockStream(Stream[object]):
    @override
    def condition(self) -> Condition:
        return MockCondition()


@pytest.fixture
def stream() -> Stream[object]:
    return MockStream()


class TestStream:
    def test_start_session(self, stream: Stream[int]) -> None:
        with stream.start_session() as session:
            stream.add_item(42)
        assert session.get_items() == [42]

    def test_start_session_with_condition(self, stream: Stream[int]) -> None:
        with stream.start_session(start_condition=lambda x: x == 2) as session:
            stream.add_item(1)
            stream.add_item(2)
            stream.add_item(3)
        assert session.get_items() == [2, 3]

    def test_wait_for(self, stream: Stream[int]) -> None:
        session = stream.start_session()

        def add_items():
            for i in range(5):
                stream.add_item(i)
                time.sleep(0.1)

        # Run adding items in a separate thread
        worker = threading.Thread(target=add_items)
        worker.start()

        # Wait until 5 items are collected
        session.wait_for(lambda numbers: len(numbers) >= 5)
        collected_numbers = session.get_items()

        assert collected_numbers == [0, 1, 2, 3, 4]  # Items should be collected

        stream.end_session(session)
        worker.join()  # Ensure thread has finished before exiting

    def test_applies_smoothing(self, stream: Stream[int]) -> None:
        stream.smoothing_fn = lambda _: 1

        with stream.start_session() as session:
            stream.add_item(42)

        assert session.get_items() == [1]



# https://github.com/Klipper3d/klipper/blob/master/klippy/extras/bed_mesh.py
from collections.abc import Iterator
from typing import Literal, TypedDict

from gcode import GCodeCommand

type _Pos = list[float]

class BedMeshError(Exception): ...

class _Params(TypedDict):
    mesh_min: tuple[float, float]
    mesh_max: tuple[float, float]
    x_count: int
    y_count: int
    mesh_x_pps: int
    mesh_y_pps: int
    algo: Literal["lagrange", "bicubic", "direct"]
    tension: float

class ZMesh:
    def __init__(self, params: _Params, name: str | None) -> None: ...
    def build_mesh(self, z_matrix: list[_Pos]) -> None: ...

class BedMesh:
    bmc: BedMeshCalibrate
    horizontal_move_z: float
    def set_mesh(self, mesh: ZMesh | None) -> None: ...

class BedMeshCalibrate:
    mesh_config: _Params
    probe_mgr: ProbeManager
    _profile_name: str
    def update_config(self, gcmd: GCodeCommand) -> None: ...
    def probe_finalize(self, offsets: list[float], positions: list[list[float]]) -> None: ...

class ProbeManager:
    def iter_rapid_path(self) -> Iterator[tuple[tuple[float, float], bool]]: ...
    def get_std_path(self) -> list[tuple[float, float]]: ...



# https://github.com/Klipper3d/klipper/blob/master/klippy/stepper.py
from typing import Literal

from mcu import MCU

type _Pos = list[float]

class MCU_stepper:
    def get_mcu(self) -> MCU: ...
    def get_name(self, short: bool = False) -> str: ...
    def get_commanded_position(self) -> _Pos: ...
    def is_active_axis(self, axis: Literal["x", "y", "z", "e"]) -> bool: ...

class PrinterRail:
    def get_steppers(self) -> list[MCU_stepper]: ...
    def get_endstops(self) -> list[tuple[object, str]]: ...



from typing import Callable, Protocol, TypeVar

from typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")


class TaskExecutor(Protocol):
    def run(self, fn: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R: ...



# https://github.com/Klipper3d/klipper/blob/master/klippy/kinematics/none.py
from typing import Protocol, TypedDict

import gcode
from stepper import MCU_stepper

# NOTE: This is a partial definition of the Kinematics class
# This module does not exist in the real Klipper
# DO NOT IMPORT THIS MODULE

type _Pos = list[float]

class Status(TypedDict):
    homed_axes: str
    axis_minimum: gcode.Coord
    axis_maximum: gcode.Coord

class Kinematics(Protocol):
    def get_steppers(self) -> list[MCU_stepper]: ...
    def get_status(self, eventtime: float) -> Status: ...
    def calc_position(self, stepper_positions: dict[str, _Pos]) -> _Pos: ...
    def clear_homing_state(
        self, axes: str | tuple[int, ...]
    ) -> None: ...  # TODO: Kalico and old klippy takes tuple[int,..]
    def note_z_not_homed(self) -> None: ...  # TODO: Kalico and old klippy has this method



from __future__ import annotations

import logging
from textwrap import dedent
from typing import TYPE_CHECKING, Callable, TypedDict, final

from cartographer.klipper.axis_twist_compensation import KlipperAxisTwistCompensationHelper
from cartographer.klipper.bed_mesh import KlipperMeshHelper
from cartographer.klipper.configuration import KlipperCartographerConfiguration
from cartographer.klipper.endstop import KlipperEndstop
from cartographer.klipper.homing import CartographerHomingChip
from cartographer.klipper.logging import setup_console_logger
from cartographer.klipper.mcu import KlipperCartographerMcu
from cartographer.klipper.mcu.mcu import Sample
from cartographer.klipper.printer import KlipperToolhead
from cartographer.klipper.probe import KlipperCartographerProbe
from cartographer.klipper.task_executor import KlipperMultiprocessingExecutor
from cartographer.klipper.temperature import PrinterTemperatureCoil
from cartographer.lib.alpha_beta_filter import AlphaBetaFilter
from cartographer.macros import ProbeAccuracyMacro, ProbeMacro, QueryProbeMacro, ZOffsetApplyProbeMacro
from cartographer.macros.axis_twist_compensation import AxisTwistCompensationMacro
from cartographer.macros.backlash import EstimateBacklashMacro
from cartographer.macros.bed_mesh import BedMeshCalibrateMacro
from cartographer.macros.scan import ScanCalibrateMacro
from cartographer.macros.touch import TouchAccuracyMacro, TouchCalibrateMacro, TouchHomeMacro, TouchMacro
from cartographer.probe import Probe, ScanMode, ScanModel, TouchMode

if TYPE_CHECKING:
    from configfile import ConfigWrapper
    from gcode import GCodeCommand

    from cartographer.printer_interface import Macro

logger = logging.getLogger(__name__)


def load_config(config: ConfigWrapper):
    pheaters = config.get_printer().load_object(config, "heaters")
    pheaters.add_sensor_factory("cartographer_coil", PrinterTemperatureCoil)
    return PrinterCartographer(config)


def smooth_with(filter: AlphaBetaFilter) -> Callable[[Sample], Sample]:
    def fn(sample: Sample) -> Sample:
        return Sample(
            sample.time,
            filter.update(measurement=sample.frequency, time=sample.time),
            sample.temperature,
            sample.position,
            sample.velocity,
        )

    return fn


class CartographerStatus(TypedDict):
    scan: CartographerScanStatus
    touch: CartographerTouchStatus


class CartographerScanStatus(TypedDict):
    current_model: str | None
    last_z_result: float | None


class CartographerTouchStatus(TypedDict):
    current_model: str | None
    last_z_result: float | None


@final
class PrinterCartographer:
    config: KlipperCartographerConfiguration

    def __init__(self, config: ConfigWrapper) -> None:
        printer = config.get_printer()
        logger.debug("Initializing Cartographer")
        self.config = KlipperCartographerConfiguration(config)
        task_executor = KlipperMultiprocessingExecutor(printer.get_reactor())

        filter = AlphaBetaFilter()
        self.mcu = KlipperCartographerMcu(config, smooth_with(filter))
        toolhead = KlipperToolhead(config, self.mcu)

        scan_config = self.config.scan_models.get("default")
        model = ScanModel(scan_config) if scan_config else None
        self.scan_mode = ScanMode(self.mcu, toolhead, self.config, model=model)
        scan_endstop = KlipperEndstop(self.mcu, self.scan_mode)

        touch_config = self.config.touch_models.get("default")
        self.touch_mode = TouchMode(self.mcu, toolhead, self.config, model=touch_config)
        probe = Probe(self.scan_mode, self.touch_mode)

        homing_chip = CartographerHomingChip(printer, scan_endstop)

        printer.lookup_object("pins").register_chip("probe", homing_chip)

        self.gcode = printer.lookup_object("gcode")
        self._configure_macro_logger()
        self.probe_macro = ProbeMacro(probe)
        self._register_macro(self.probe_macro)
        self._register_macro(ProbeAccuracyMacro(probe, toolhead))
        query_probe_macro = QueryProbeMacro(probe)
        self._register_macro(query_probe_macro)

        self._register_macro(ZOffsetApplyProbeMacro(probe, toolhead))

        self.touch_macro = TouchMacro(self.touch_mode)
        self._register_macro(self.touch_macro)
        self._register_macro(TouchAccuracyMacro(self.touch_mode, toolhead))
        touch_home = TouchHomeMacro(self.touch_mode, toolhead, self.config.zero_reference_position)
        self._register_macro(touch_home)

        self._register_macro(
            BedMeshCalibrateMacro(
                probe,
                toolhead,
                KlipperMeshHelper(config, self.gcode),
                task_executor,
                self.config,
            )
        )

        self._register_macro(ScanCalibrateMacro(probe, toolhead, self.config))
        self._register_macro(TouchCalibrateMacro(self.touch_mode, toolhead, self.config))

        self._register_macro(
            AxisTwistCompensationMacro(probe, toolhead, KlipperAxisTwistCompensationHelper(config), self.config)
        )
        self._register_macro(EstimateBacklashMacro(toolhead, self.scan_mode))

        printer.add_object(
            "probe",
            KlipperCartographerProbe(
                toolhead,
                self.scan_mode,
                self.probe_macro,
                query_probe_macro,
            ),
        )

    def _register_macro(self, macro: Macro[GCodeCommand]) -> None:
        self.gcode.register_command(macro.name, catch_macro_errors(macro.run), desc=macro.description)

    def _configure_macro_logger(self) -> None:
        handler = setup_console_logger(self.gcode)

        log_level = logging.DEBUG if self.config.verbose else logging.INFO
        handler.setLevel(log_level)

    def get_status(self, eventtime: float) -> CartographerStatus:
        del eventtime
        return CartographerStatus(
            scan=CartographerScanStatus(
                current_model=self.scan_mode.model.name if self.scan_mode.model else None,
                last_z_result=self.probe_macro.last_trigger_position,
            ),
            touch=CartographerTouchStatus(
                current_model=self.touch_mode.model.name if self.touch_mode.model else None,
                last_z_result=self.touch_macro.last_trigger_position,
            ),
        )


def catch_macro_errors(func: Callable[[GCodeCommand], None]) -> Callable[[GCodeCommand], None]:
    def wrapper(gcmd: GCodeCommand) -> None:
        try:
            return func(gcmd)
        except RuntimeError as e:
            raise gcmd.error(dedent(str(e)).replace("\n", " ").strip()) from e

    return wrapper



from __future__ import annotations

from typing import TYPE_CHECKING, final

import pytest
from typing_extensions import TypeAlias

from cartographer.printer_interface import HomingState, Mcu, Position, Sample, TemperatureStatus, Toolhead
from cartographer.probe.touch_mode import Configuration, TouchMode

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from cartographer.configuration import TouchModelConfiguration


Probe: TypeAlias = TouchMode[object]


@final
class MockConfiguration(Configuration):
    move_speed = 42.0

    touch_samples = 5
    touch_max_samples = 10
    x_offset = 10.0
    y_offset = 10.0
    mesh_min = (10, 10)
    mesh_max = (100, 100)


class MockModel:
    name: str = "default"
    threshold: int = 5
    speed: float = 10.0
    z_offset: float = 0.0

    def save_z_offset(self, new_offset: float) -> None:
        del new_offset
        pass


@pytest.fixture
def config() -> Configuration:
    return MockConfiguration()


@pytest.fixture
def model() -> TouchModelConfiguration:
    return MockModel()


@pytest.fixture
def probe(mcu: Mcu[object, Sample], toolhead: Toolhead, config: Configuration, model: TouchModelConfiguration) -> Probe:
    return Probe(mcu, toolhead, config, model=model)


@pytest.fixture
def homing_state(mocker: MockerFixture, probe: Probe) -> HomingState:
    mock = mocker.Mock(spec=HomingState, autospec=True)
    mock.endstops = [probe]
    return mock


def test_probe_success(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.z_homing_move = mocker.Mock(return_value=0.5)
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, 1))

    assert probe.perform_probe() == 0.5


def test_probe_moves_below_5(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.z_homing_move = mocker.Mock(return_value=0.5)
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, 1))
    move_spy = mocker.spy(toolhead, "move")

    _ = probe.perform_probe()

    assert move_spy.mock_calls[0] == mocker.call(z=5, speed=mocker.ANY)


def test_does_not_move_above_5(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.z_homing_move = mocker.Mock(return_value=0.5)
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, 10))
    move_spy = mocker.spy(toolhead, "move")

    _ = probe.perform_probe()

    assert move_spy.mock_calls[0] != mocker.call(z=5, speed=mocker.ANY)


def test_probe_standard_deviation_failure(
    mocker: MockerFixture, toolhead: Toolhead, probe: Probe, config: MockConfiguration
) -> None:
    config.touch_max_samples = 5
    toolhead.z_homing_move = mocker.Mock(side_effect=[1.000, 1.002, 1.1, 1.016, 1.018])
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, 1))

    with pytest.raises(RuntimeError, match="unable to find"):
        _ = probe.perform_probe()


def test_probe_suceeds_on_more(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.z_homing_move = mocker.Mock(side_effect=[1.0, 1.01, 1.5, 0.5, 0.5, 0.5, 0.5, 0.5])
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, 1))

    assert probe.perform_probe() == 0.5


def test_probe_suceeds_on_spread_samples(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.z_homing_move = mocker.Mock(side_effect=[0.5, 1.0, 1.5, 0.5, 2.5, 0.5, 3.5, 0.5, 4.5, 0.5])
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, 1))

    assert probe.perform_probe() == 0.5


def test_probe_unhomed_z(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.is_homed = mocker.Mock(return_value=False)

    with pytest.raises(RuntimeError, match="z axis must be homed"):
        _ = probe.perform_probe()


def test_home_start_invalid_threshold(model: TouchModelConfiguration, probe: Probe) -> None:
    model.threshold = 0

    with pytest.raises(RuntimeError, match="threshold must be greater than 0"):
        _ = probe.home_start(print_time=0.0)


def test_home_wait(mocker: MockerFixture, mcu: Mcu[object, Sample], probe: Probe) -> None:
    mcu.stop_homing = mocker.Mock(return_value=1.5)

    assert probe.home_wait(home_end_time=1.0) == 1.5


def test_on_home_end(mocker: MockerFixture, probe: Probe, homing_state: HomingState) -> None:
    homed_position_spy = mocker.spy(homing_state, "set_z_homed_position")

    probe.on_home_end(homing_state)

    assert homed_position_spy.called == 1


def test_abort_if_current_extruder_too_hot(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.get_extruder_temperature = mocker.Mock(return_value=TemperatureStatus(156, 0))

    with pytest.raises(RuntimeError, match="nozzle temperature must be below 150C"):
        _ = probe.home_start(print_time=0.0)


def test_abort_if_current_extruder_target_too_hot(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.get_extruder_temperature = mocker.Mock(return_value=TemperatureStatus(0, 156))

    with pytest.raises(RuntimeError, match="nozzle temperature must be below 150C"):
        _ = probe.home_start(print_time=0.0)


def test_nozzle_outside_bounds(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, 1))

    with pytest.raises(RuntimeError, match="outside .* boundaries"):
        _ = probe.home_start(0)


def test_probe_outside_bounds(mocker: MockerFixture, toolhead: Toolhead, probe: Probe) -> None:
    toolhead.get_position = mocker.Mock(return_value=Position(95, 95, 1))

    with pytest.raises(RuntimeError, match="outside .* boundaries"):
        _ = probe.home_start(0)



# https://github.com/Klipper3d/klipper/blob/master/klippy/gcode.py
from typing import Callable, NamedTuple, overload

class CommandError(Exception): ...

class GCodeCommand:
    error: type[CommandError]

    def respond_raw(self, msg: str) -> None: ...
    def respond_info(self, msg: str, log: bool = True) -> None: ...
    def get_command_parameters(self) -> dict[str, str]: ...
    @overload
    def get(
        self,
        name: str,
        default: str = ...,
    ) -> str: ...
    @overload
    def get(
        self,
        name: str,
        default: None,
    ) -> str | None: ...
    @overload
    def get_int(
        self,
        name: str,
        default: int = ...,
        minval: int | None = None,
        maxval: int | None = None,
    ) -> int: ...
    @overload
    def get_int(
        self,
        name: str,
        default: None,
        minval: int | None = None,
        maxval: int | None = None,
    ) -> int | None: ...
    @overload
    def get_float(
        self,
        name: str,
        default: float = ...,
        minval: float | None = None,
        maxval: float | None = None,
        above: float | None = None,
        below: float | None = None,
    ) -> float: ...
    @overload
    def get_float(
        self,
        name: str,
        default: None,
        minval: float | None = None,
        maxval: float | None = None,
        above: float | None = None,
        below: float | None = None,
    ) -> float | None: ...

class Coord(NamedTuple):
    x: float
    y: float
    z: float
    e: float

class GCodeDispatch:
    error: type[CommandError]
    Coord: type[Coord]

    def respond_raw(self, msg: str) -> None: ...
    def respond_info(self, msg: str, log: bool = True) -> None: ...
    def run_script_from_command(self, script: str) -> None: ...
    @overload
    def register_command(
        self,
        cmd: str,
        func: Callable[[GCodeCommand], None],
        when_not_ready: bool = False,
        desc: str | None = None,
    ) -> None: ...
    @overload
    def register_command(
        self,
        cmd: str,
        func: None,
        when_not_ready: bool = False,
        desc: str | None = None,
    ) -> Callable[[GCodeCommand], None] | None: ...
    def create_gcode_command(
        self,
        command: str,
        commandline: str,
        params: dict[str, str],
    ) -> GCodeCommand: ...



# Helper script to determine a Z height
from typing import Callable

from gcode import GCodeCommand
from klippy import Printer

type _Pos = list[float]

def verify_no_manual_probe(printer: Printer) -> None: ...

class ManualProbeHelper:
    def __init__(
        self,
        printer: Printer,
        gcmd: GCodeCommand,
        finalize_callback: Callable[[_Pos | None], None],
    ) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING, overload

import pytest
from pytest_bdd import given, parsers, then
from typing_extensions import override

from cartographer.printer_interface import MacroParams

if TYPE_CHECKING:
    from pytest import LogCaptureFixture


@pytest.fixture
def params() -> MacroParams:
    return MockParams()


class MockParams(MacroParams):
    def __init__(self) -> None:
        self.params: dict[str, str] = {}

    @override
    def get(self, name: str, default: str = ...) -> str:
        return str(self.params.get(name, default))

    @overload
    def get_float(self, name: str, default: float = ..., *, above: float = ..., minval: float = ...) -> float: ...
    @overload
    def get_float(self, name: str, default: None, *, above: float = ..., minval: float = ...) -> float | None: ...

    @override
    def get_float(
        self, name: str, default: float | None = ..., *, above: float = ..., minval: float = ...
    ) -> float | None:
        opt = self.params.get(name, default)
        return float(opt) if opt is not None else None

    @override
    def get_int(self, name: str, default: int = ..., *, minval: int = ..., maxval: int = ...) -> int:
        return int(self.params.get(name, default))


@given("macro parameters:")
def given_parameters(datatable: list[list[str]], params: MockParams):
    params.params = {key: value for key, value in datatable}


@then(parsers.parse('it should log "{output}"'))
def then_log_result(caplog: LogCaptureFixture, output: str):
    assert output in caplog.text



# https://github.com/Klipper3d/klipper/blob/master/klippy/kinematics/extruder.py
from extras.heaters import Heater

class Extruder:
    def get_name(self) -> str: ...
    def get_heater(self) -> Heater: ...



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

from extras.probe import PrinterProbe
from typing_extensions import override

from cartographer.klipper.utils import reraise_as_command_error

if TYPE_CHECKING:
    from gcode import GCodeCommand

    from cartographer.klipper.printer import KlipperToolhead
    from cartographer.macros.probe import ProbeMacro, QueryProbeMacro
    from cartographer.printer_interface import ProbeMode


class ProbeStatus(TypedDict):
    name: str
    last_query: int
    last_z_result: float


class ProbeParams(TypedDict):
    probe_speed: float
    lift_speed: float


# TODO: Get the values from some configuration?
DEFAULT_LIFT_SPEED = 10
DEFAULT_PROBE_SPEED = 3


class KlipperProbeSession:
    def __init__(self, probe: ProbeMode, toolhead: KlipperToolhead) -> None:
        self._probe: ProbeMode = probe
        self._results: list[list[float]] = []
        self.toolhead: KlipperToolhead = toolhead

    @reraise_as_command_error
    def run_probe(self, gcmd: GCodeCommand) -> None:
        del gcmd
        pos = self.toolhead.get_position()
        trigger_pos = self._probe.perform_probe()
        self._results.append([pos.x, pos.y, trigger_pos])

    def pull_probed_results(self):
        result = self._results
        self._results = []
        return result

    def end_probe_session(self) -> None:
        pass


class KlipperCartographerProbe(PrinterProbe):
    def __init__(
        self,
        toolhead: KlipperToolhead,
        probe: ProbeMode,
        probe_macro: ProbeMacro,
        query_probe_macro: QueryProbeMacro,
    ) -> None:
        self.probe: ProbeMode = probe
        self.probe_macro: ProbeMacro = probe_macro
        self.query_probe_macro: QueryProbeMacro = query_probe_macro
        self.toolhead: KlipperToolhead = toolhead
        self.lift_speed: float = DEFAULT_LIFT_SPEED

    @override
    def get_probe_params(self, gcmd: GCodeCommand | None = None) -> ProbeParams:
        if gcmd is None:
            return ProbeParams(lift_speed=self.lift_speed, probe_speed=DEFAULT_PROBE_SPEED)

        lift_speed = gcmd.get_float("LIFT_SPEED", default=self.lift_speed, above=0)
        probe_speed = gcmd.get_float("SPEED", default=DEFAULT_PROBE_SPEED, above=0)
        return ProbeParams(lift_speed=lift_speed, probe_speed=probe_speed)

    @override
    def get_offsets(self) -> tuple[float, float, float]:
        return self.probe.offset.as_tuple()

    @override
    def get_status(self, eventtime: float) -> ProbeStatus:
        return ProbeStatus(
            name="cartographer",
            last_query=1 if self.query_probe_macro.last_triggered else 0,
            last_z_result=self.probe_macro.last_trigger_position or 0,
        )

    @override
    def start_probe_session(self, gcmd: GCodeCommand) -> KlipperProbeSession:
        return KlipperProbeSession(self.probe, self.toolhead)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import NDArray


def compute_mad(samples: NDArray[np.float64] | Sequence[float]) -> float:
    samples = np.asarray(samples)
    if len(samples) < 1:
        return float("inf")
    median = np.median(samples)
    mad = np.median(np.abs(samples - median))
    return float(mad)



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import TYPE_CHECKING

import pytest
from numpy.polynomial import Polynomial

from cartographer.printer_interface import Position, Sample
from cartographer.probe.scan_model import ScanModel

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from cartographer.configuration import ScanModelConfiguration


@dataclass
class MockSample(Sample):
    time: float
    frequency: float
    position: Position | None
    velocity: float | None = None


class MockConfiguration:
    name: str = "default"
    coefficients: list[float] = [1 / 3.0] * 9
    domain: tuple[float, float] = (1, 3)
    z_offset: float = 0

    def save_z_offset(self, new_offset: float) -> None:
        self.z_offset = new_offset


@pytest.fixture
def config() -> ScanModelConfiguration:
    return MockConfiguration()


@pytest.fixture
def model(mocker: MockerFixture, config: MockConfiguration) -> ScanModel:
    def eval_poly(x: float) -> float:
        return x

    poly = mocker.Mock(spec=Polynomial, autospec=True, side_effect=eval_poly)
    poly.domain = config.domain

    model = ScanModel(config)
    return model


def test_fit() -> None:
    samples = [MockSample(time=i, frequency=1 / i, position=Position(0, 0, 0)) for i in range(1, 20)]

    fit = ScanModel.fit(samples)

    assert fit.domain[0] == 1
    assert fit.domain[1] == 19


def test_from_config(config: ScanModelConfiguration) -> None:
    model = ScanModel(config)
    assert isinstance(model, ScanModel)
    assert model.poly.domain[0] == config.domain[0]
    assert model.poly.domain[1] == config.domain[1]


def test_frequency_to_distance(model: ScanModel) -> None:
    frequency = 1 / 3.0
    distance = model.frequency_to_distance(frequency)
    assert isinstance(distance, float)
    assert distance != math.inf and distance != -math.inf


def test_distance_to_frequency(model: ScanModel) -> None:
    distance = 2.5
    frequency = model.distance_to_frequency(distance)
    assert isinstance(frequency, float)
    assert frequency > 0


def test_distance_to_frequency_out_of_range(model: ScanModel) -> None:
    with pytest.raises(RuntimeError, match="attempted to map out-of-range distance"):
        _ = model.distance_to_frequency(11)  # Out of z_range


def test_frequency_to_distance_applies_offset(model: ScanModel, config: ScanModelConfiguration) -> None:
    config.z_offset = -0.5
    frequency = 1 / 3.0

    distance = model.frequency_to_distance(frequency)

    assert distance == 2.5


def test_distance_to_frequency_applies_offset(model: ScanModel, config: ScanModelConfiguration) -> None:
    config.z_offset = -0.5
    distance = 2.5

    frequency = model.distance_to_frequency(distance)

    assert frequency == pytest.approx(1 / 3)  # pyright: ignore[reportUnknownMemberType]


def test_frequency_to_distance_out_of_range(model: ScanModel) -> None:
    high_frequency_dist = model.frequency_to_distance(1000000)  # Out of z_range
    low_frequency_dist = model.frequency_to_distance(0.1)  # Out of z_range

    assert low_frequency_dist == float("inf")
    assert high_frequency_dist == float("-inf")



from functools import wraps
from typing import Callable, TypeVar

from gcode import CommandError
from typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")


def reraise_as_command_error(func: Callable[P, R]) -> Callable[P, R]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        try:
            return func(*args, **kwargs)
        except RuntimeError as e:
            raise CommandError(str(e)) from e

    return wrapper



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Protocol, final

import numpy as np
from typing_extensions import override

from cartographer.configuration import TouchModelConfiguration
from cartographer.lib.statistics import compute_mad
from cartographer.printer_interface import Macro, MacroParams
from cartographer.probe.touch_mode import STD_TOLERANCE, TouchMode

if TYPE_CHECKING:
    from cartographer.printer_interface import Toolhead


logger = logging.getLogger(__name__)

Probe = TouchMode[object]

logger = logging.getLogger(__name__)


class Configuration(Protocol):
    zero_reference_position: tuple[float, float]
    touch_samples: int
    move_speed: float

    def save_new_touch_model(self, name: str, speed: float, threshold: int) -> TouchModelConfiguration: ...


@final
class TouchMacro(Macro[MacroParams]):
    name = "TOUCH"
    description = "Touch the bed to get the height offset at the current position."
    last_trigger_position: float | None = None

    def __init__(self, probe: Probe) -> None:
        self._probe = probe

    @override
    def run(self, params: MacroParams) -> None:
        trigger_position = self._probe.perform_probe()
        logger.info("Result is z=%.6f", trigger_position)
        self.last_trigger_position = trigger_position


@final
class TouchAccuracyMacro(Macro[MacroParams]):
    name = "TOUCH_ACCURACY"
    description = "Touch the bed multiple times to measure the accuracy of the probe."

    def __init__(self, probe: Probe, toolhead: Toolhead) -> None:
        self._probe = probe
        self._toolhead = toolhead

    @override
    def run(self, params: MacroParams) -> None:
        lift_speed = params.get_float("LIFT_SPEED", 5.0, above=0)
        retract = params.get_float("SAMPLE_RETRACT_DIST", 1.0, minval=0)
        sample_count = params.get_int("SAMPLES", 5, minval=1)
        position = self._toolhead.get_position()

        logger.info(
            "TOUCH_ACCURACY at X:%.3f Y:%.3f Z:%.3f (samples=%d retract=%.3f lift_speed=%.1f)",
            position.x,
            position.y,
            position.z,
            sample_count,
            retract,
            lift_speed,
        )

        self._toolhead.move(z=position.z + retract, speed=lift_speed)
        measurements: list[float] = []
        while len(measurements) < sample_count:
            trigger_pos = self._probe.perform_probe()
            measurements.append(trigger_pos)
            pos = self._toolhead.get_position()
            self._toolhead.move(z=pos.z + retract, speed=lift_speed)
        logger.debug("Measurements gathered: %s", ", ".join(f"{m:.6f}" for m in measurements))

        max_value = max(measurements)
        min_value = min(measurements)
        range_value = max_value - min_value
        avg_value = np.mean(measurements)
        median = np.median(measurements)
        std_dev = np.std(measurements)
        mad = compute_mad(measurements)

        logger.info(
            """
            touch accuracy results: maximum %.6f, minimum %.6f, range %.6f,
            average %.6f, median %.6f, standard deviation %.6f
            median absolute deviation %.6f
            """,
            max_value,
            min_value,
            range_value,
            avg_value,
            median,
            std_dev,
            mad,
        )


@final
class TouchHomeMacro(Macro[MacroParams]):
    name = "TOUCH_HOME"
    description = "Touch the bed to home Z axis"

    def __init__(
        self,
        probe: Probe,
        toolhead: Toolhead,
        home_position: tuple[float, float],
    ) -> None:
        self._probe = probe
        self._toolhead = toolhead
        self._home_position = home_position

    @override
    def run(self, params: MacroParams) -> None:
        if not self._toolhead.is_homed("x") or not self._toolhead.is_homed("y"):
            msg = "must home x and y before touch homing"
            raise RuntimeError(msg)

        self._toolhead.move(
            x=self._home_position[0],
            y=self._home_position[1],
            speed=self._probe.config.move_speed,
        )
        self._toolhead.wait_moves()

        forced_z = False
        if not self._toolhead.is_homed("z"):
            forced_z = True
            _, z_max = self._toolhead.get_z_axis_limits()
            self._toolhead.set_z_position(z=z_max - 10)

        try:
            trigger_pos = self._probe.perform_probe()
        finally:
            if forced_z:
                self._toolhead.clear_z_homing_state()

        pos = self._toolhead.get_position()
        self._toolhead.set_z_position(pos.z - (trigger_pos - self._probe.offset.z))
        logger.info(
            "Touch home at (%.3f,%.3f) adjusted z by %.3f, offset %.3f",
            pos.x,
            pos.y,
            -trigger_pos,
            -self._probe.offset.z,
        )


@final
class CalibrationModel(TouchModelConfiguration):
    name = "calibration"
    z_offset = 0.0

    def __init__(self, *, speed: float, threshold: int) -> None:
        self.speed = speed
        self.threshold = threshold

    @override
    def save_z_offset(self, new_offset: float) -> None:
        msg = "calibration model cannot be saved"
        raise RuntimeError(msg)


SAFE_TRIGGER_MIN_HEIGHT = -0.2  # Initial home too far
MAD_TOLERANCE = STD_TOLERANCE * 0.6745  # Convert std to mad

MIN_ALLOWED_STEP = 75
MAX_ALLOWED_STEP = 500


def compute_step_increase(current_threshold: int, score: float) -> int:
    # Compute max allowed step from threshold
    allowed_step = (current_threshold * 0.06) ** 1.15
    allowed_max_step = max(MIN_ALLOWED_STEP, min(MAX_ALLOWED_STEP, allowed_step))

    # Compute step directly based on how bad the score is
    ratio = max(1.0, score / MAD_TOLERANCE)
    raw_step = allowed_max_step * (1 - (1 / ratio))

    # Clamp and return
    step = int(round(min(allowed_max_step, max(MIN_ALLOWED_STEP, raw_step))))
    return step


@final
class TouchCalibrateMacro(Macro[MacroParams]):
    name = "TOUCH_CALIBRATE"
    description = "Run the touch calibration"

    def __init__(self, probe: Probe, toolhead: Toolhead, config: Configuration) -> None:
        self._probe = probe
        self._toolhead = toolhead
        self._config = config

    @override
    def run(self, params: MacroParams) -> None:
        name = params.get("MODEL_NAME", "default")
        speed = params.get_int("SPEED", default=3, minval=1, maxval=5)
        threshold_start = params.get_int("START", default=500, minval=100)
        threshold_max = params.get_int("MAX", default=3000, minval=threshold_start)

        if not self._toolhead.is_homed("x") or not self._toolhead.is_homed("y"):
            msg = "must home x and y before calibration"
            raise RuntimeError(msg)

        self._toolhead.move(
            x=self._config.zero_reference_position[0],
            y=self._config.zero_reference_position[1],
            speed=self._config.move_speed,
        )
        self._toolhead.wait_moves()

        forced_z = False
        if not self._toolhead.is_homed("z"):
            forced_z = True
            _, z_max = self._toolhead.get_z_axis_limits()
            self._toolhead.set_z_position(z=z_max - 10)

        logger.info(
            "Starting touch calibration at speed %d, starting threshold %d, max threshold %d",
            speed,
            threshold_start,
            threshold_max,
        )
        try:
            threshold = self._find_acceptable_threshold(threshold_start, threshold_max, speed)
        finally:
            if forced_z:
                self._toolhead.clear_z_homing_state()

        if threshold is None:
            logger.info(
                """
                Failed to calibrate touch with thresholds between %d and %d.
                If you want to continue calibration, increase the max threshold.
                E.g. %s START=%d MAX=%d
                """,
                threshold_start,
                threshold_max,
                self.name,
                threshold_max,
                threshold_max + 1000,
            )
            return

        logger.info("Touch calibrated at speed %d, threshold %d", speed, threshold)
        self._probe.model = self._config.save_new_touch_model(name, speed, threshold)
        logger.info(
            """
            Touch model %s has been saved
            for the current session. The SAVE_CONFIG command will
            update the printer config file and restart the printer.
            """,
            name,
        )

    def _find_acceptable_threshold(self, threshold_start: int, threshold_max: int, speed: int) -> int | None:
        current_threshold = threshold_start
        while current_threshold <= threshold_max:
            model = CalibrationModel(speed=speed, threshold=current_threshold)
            score, samples = self._evaluate_threshold(model)

            logger.info(
                "Threshold %d score: %.6f",
                current_threshold,
                score,
            )
            logger.debug(
                "Threshold %d samples: %s",
                current_threshold,
                ", ".join(f"{s:.6f}" for s in samples),
            )

            # Accept the first threshold within tolerance
            if score <= MAD_TOLERANCE:
                logger.info("Threshold %d with score %.6f is acceptable.", current_threshold, score)
                return current_threshold

            step = compute_step_increase(current_threshold, score)
            current_threshold += step
            logger.info("Next threshold %d (+%d)", current_threshold, step)

        return None  # No acceptable threshold found

    def _evaluate_threshold(self, model: CalibrationModel) -> tuple[float, list[float]]:
        old_model = self._probe.model
        try:
            self._probe.model = model
            samples: list[float] = []

            for _ in range(self._config.touch_samples * 3):
                pos = self._probe.perform_single_probe()
                if pos < SAFE_TRIGGER_MIN_HEIGHT:
                    msg = f"""
                        probe triggered at {pos:.3f}, far below expected bed level
                        Please ensure that z is homed to within {SAFE_TRIGGER_MIN_HEIGHT:.1f}mm of the bed.
                    """
                    raise RuntimeError(msg)
                samples.append(pos)

                if len(samples) >= 3:
                    current_score = self._evaluate_samples(samples)
                    # Early exit only for clearly unstable thresholds
                    if current_score > 5 * MAD_TOLERANCE:
                        return current_score, samples

            final_score = self._evaluate_samples(samples)
            return final_score, samples
        finally:
            self._probe.model = old_model

    def _evaluate_samples(self, samples: list[float]) -> float:
        return compute_mad(samples)



# https://github.com/Klipper3d/klipper/blob/master/klippy/pins.py

from typing import Protocol, TypedDict

from mcu import MCU_endstop

class error(Exception): ...

class _PinParams(TypedDict):
    chip: _Chip
    chip_name: str
    pin: str
    invert: int
    pullup: int

class _Chip(Protocol):
    def setup_pin(self, pin_type: str, pin_params: _PinParams) -> MCU_endstop: ...

class PrinterPins:
    error: type[error]
    def register_chip(self, chip_name: str, chip: _Chip) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Protocol, TypeVar, final

import greenlet
from typing_extensions import override

from cartographer.stream import Condition, Session, Stream

if TYPE_CHECKING:
    from reactor import Reactor


@final
class KlipperCondition(Condition):
    """The Klipper equivalent of [threading.Condition](https://docs.python.org/3/library/threading.html#condition-objects)"""

    def __init__(self, reactor: Reactor):
        self.reactor = reactor
        self.waiting: list[greenlet.greenlet] = []

    @override
    def notify_all(self):
        for wait in self.waiting:
            self.reactor.update_timer(wait.timer, self.reactor.NOW)

    @override
    def wait_for(self, predicate: Callable[[], bool]) -> None:
        wait = greenlet.getcurrent()
        self.waiting.append(wait)
        while True:
            if predicate():
                break
            _ = self.reactor.pause(self.reactor.NEVER)
        self.waiting.remove(wait)


T = TypeVar("T")


class KlipperStreamMcu(Protocol):
    def start_streaming(self) -> None:
        """Used to ask the MCU to start sending data."""
        ...

    def stop_streaming(self) -> None:
        """Stop the MCU from sending data.
        Will be called when the last session ends.
        """
        ...


@final
class KlipperStream(Stream[T]):
    def __init__(
        self,
        mcu: KlipperStreamMcu,
        reactor: Reactor,
        smoothing_fn: Callable[[T], T] | None = None,
    ):
        self.reactor = reactor
        self.mcu = mcu
        super().__init__(smoothing_fn)

    @override
    def condition(self) -> Condition:
        return KlipperCondition(self.reactor)

    @override
    def start_session(self, start_condition: Callable[[T], bool] | None = None) -> Session[T]:
        if len(self.sessions) == 0:
            self.mcu.start_streaming()
        return super().start_session(start_condition)

    @override
    def end_session(self, session: Session[T]) -> None:
        super().end_session(session)
        if len(self.sessions) == 0:
            self.mcu.stop_streaming()



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import pytest
from typing_extensions import TypeAlias

from cartographer.macros.bed_mesh import BedMeshCalibrateMacro, Configuration, MeshHelper, MeshPoint
from cartographer.printer_interface import MacroParams, Position, Toolhead
from cartographer.probe import Probe, ScanMode
from cartographer.probe.scan_mode import Model
from cartographer.probe.touch_mode import TouchMode

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from cartographer.interfaces import TaskExecutor
    from cartographer.stream import Session


@dataclass
class Sample:
    time: float
    frequency: float
    position: Position | None = field(default_factory=lambda: Position(10, 10, 10))
    velocity: float | None = None


Macro: TypeAlias = BedMeshCalibrateMacro[MacroParams]
Helper: TypeAlias = MeshHelper[MacroParams]


@pytest.fixture
def scan_model(mocker: MockerFixture) -> Model:
    return mocker.MagicMock(spec=Model, autospec=True, instance=True)


@pytest.fixture
def offset() -> Position:
    return Position(0, 0, 0)


@pytest.fixture
def scan_mode(
    mocker: MockerFixture, scan_model: Model, session: Session[Sample], offset: Position
) -> ScanMode[object, Sample]:
    scan_mode = mocker.MagicMock(spec=ScanMode, autospec=True, instance=True)
    scan_mode.model = scan_model
    scan_mode.offset = offset
    scan_mode.probe_height = 10.0
    scan_mode.start_session = mocker.Mock(return_value=session)
    return scan_mode


@pytest.fixture
def touch_mode(mocker: MockerFixture) -> TouchMode[object]:
    touch_mode = mocker.MagicMock(spec=TouchMode, autospec=True, instance=True)
    return touch_mode


@pytest.fixture
def probe(scan_mode: ScanMode[object, Sample], touch_mode: TouchMode[object]) -> Probe:
    return Probe(scan_mode, touch_mode)


@pytest.fixture
def helper(mocker: MockerFixture) -> Helper:
    return mocker.MagicMock(spec=MeshHelper, autospec=True, instance=True)


class MockConfiguration(Configuration):
    scan_speed: float = 400
    scan_mesh_runs: int = 1
    scan_height: float = 5.0


@pytest.fixture
def config() -> Configuration:
    return MockConfiguration()


@pytest.fixture
def macro(
    probe: Probe,
    toolhead: Toolhead,
    helper: Helper,
    task_executor: TaskExecutor,
    config: Configuration,
) -> Macro:
    return BedMeshCalibrateMacro(probe, toolhead, helper, task_executor, config)


def test_run_valid_scan(mocker: MockerFixture, macro: Macro, toolhead: Toolhead, helper: Helper):
    helper.prepare_scan_path = mocker.Mock(return_value=[MeshPoint(10, 10, False), MeshPoint(20, 20, False)])
    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="scan")
    params.get_float = mocker.Mock(return_value=42.0)
    move_spy = mocker.spy(toolhead, "move")
    finalize_spy = mocker.spy(helper, "finalize")

    macro.run(params)

    assert [
        mocker.call(x=10, y=10, speed=42.0),
        mocker.call(x=20, y=20, speed=42.0),
    ] in move_spy.mock_calls
    assert finalize_spy.call_count == 1


def test_applies_offsets(
    mocker: MockerFixture,
    macro: Macro,
    toolhead: Toolhead,
    helper: Helper,
    offset: Position,
):
    helper.prepare_scan_path = mocker.Mock(return_value=[MeshPoint(10, 10, False), MeshPoint(20, 20, False)])
    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="scan")
    params.get_float = mocker.Mock(return_value=42.0)
    offset.x = -5
    offset.y = 5
    move_spy = mocker.spy(toolhead, "move")

    macro.run(params)

    assert [
        mocker.call(x=15, y=5, speed=42.0),
        mocker.call(x=25, y=15, speed=42.0),
    ] in move_spy.mock_calls


def test_multiple_runs(
    mocker: MockerFixture,
    macro: Macro,
    toolhead: Toolhead,
    helper: Helper,
):
    helper.prepare_scan_path = mocker.Mock(return_value=[MeshPoint(10, 10, False), MeshPoint(20, 20, False)])
    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="scan")
    params.get_int = mocker.Mock(return_value=3)
    params.get_float = mocker.Mock(return_value=42.0)
    move_spy = mocker.spy(toolhead, "move")

    macro.run(params)

    assert [
        # first
        mocker.call(x=10, y=10, speed=42.0),
        mocker.call(x=20, y=20, speed=42.0),
        # second
        mocker.call(x=20, y=20, speed=42.0),
        mocker.call(x=10, y=10, speed=42.0),
        # third
        mocker.call(x=10, y=10, speed=42.0),
        mocker.call(x=20, y=20, speed=42.0),
    ] in move_spy.mock_calls


def test_run_invalid_method(mocker: MockerFixture, macro: Macro, helper: Helper):
    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="invalid")
    orig_macro_spy = mocker.spy(helper, "orig_macro")
    finalize_spy = mocker.spy(helper, "finalize")

    macro.run(params)
    assert orig_macro_spy.call_count == 1
    assert finalize_spy.call_count == 0


def test_calculate_positions_no_valid_samples(
    mocker: MockerFixture,
    macro: Macro,
    scan_model: Model,
    helper: Helper,
    session: Session[Sample],
):
    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="scan")
    helper.prepare_scan_path = mocker.Mock(return_value=[MeshPoint(10, 10, True)])
    scan_model.frequency_to_distance = mocker.Mock(return_value=float("nan"))
    session.get_items = lambda: [Sample(time=0.0, frequency=100) for _ in range(2)]

    with pytest.raises(RuntimeError, match="no valid samples"):
        macro.run(params)


def test_calculate_positions_cluster_no_samples(
    mocker: MockerFixture,
    macro: Macro,
    scan_model: Model,
    helper: Helper,
    session: Session[Sample],
):
    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="scan")
    helper.prepare_scan_path = mocker.Mock(return_value=[MeshPoint(10, 10, True), MeshPoint(20, 20, True)])
    scan_model.frequency_to_distance = mocker.Mock(return_value=1)
    session.get_items = lambda: [Sample(time=0.0, frequency=100) for _ in range(2)]

    with pytest.raises(RuntimeError, match="no samples"):
        macro.run(params)


def test_finalize_with_positions(
    mocker: MockerFixture,
    macro: Macro,
    scan_mode: ScanMode[object, Sample],
    scan_model: Model,
    helper: Helper,
    session: Session[Sample],
):
    scan_height = 5

    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="scan")
    params.get_float = mocker.Mock(return_value=scan_height)
    helper.prepare_scan_path = mocker.Mock(return_value=[MeshPoint(10, 10, True), MeshPoint(20, 20, True)])
    session.get_items = lambda: [
        Sample(time=0.0, frequency=1, position=Position(10 + 10 * i, 10 + 10 * i, 5)) for i in range(2)
    ]

    distances = [1, 2]
    positions = [Position(10, 10, scan_height), Position(20, 20, scan_height)]
    scan_model.frequency_to_distance = mocker.Mock(side_effect=distances)

    finalize_spy = mocker.spy(helper, "finalize")
    expected_positions = [
        Position(pos.x, pos.y, scan_height + scan_mode.probe_height - dist) for pos, dist in zip(positions, distances)
    ]

    macro.run(params)

    assert finalize_spy.mock_calls == [mocker.call(Position(0, 0, 0), expected_positions)]


def test_probe_applies_axis_twist_compensation(
    mocker: MockerFixture,
    macro: Macro,
    scan_model: Model,
    toolhead: Toolhead,
    helper: Helper,
    session: Session[Sample],
):
    z_comp = 0.5
    distances = [1, 2]
    positions = [Position(10, 10, 5), Position(20, 20, 5)]

    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="scan")
    helper.prepare_scan_path = mocker.Mock(return_value=[MeshPoint(10, 10, True), MeshPoint(20, 20, True)])
    session.get_items = lambda: [
        Sample(time=0.0, frequency=1, position=Position(10 + 10 * i, 10 + 10 * i, 5)) for i in range(2)
    ]
    scan_model.frequency_to_distance = mocker.Mock(side_effect=distances)
    toolhead.apply_axis_twist_compensation = lambda position: Position(position.x, position.y, z_comp)

    finalize_spy = mocker.spy(helper, "finalize")
    expected_positions = [Position(pos.x, pos.y, z_comp) for pos in positions]

    macro.run(params)

    assert finalize_spy.mock_calls == [mocker.call(Position(0, 0, 0), expected_positions)]


def test_ignores_unincluded_points(
    mocker: MockerFixture,
    scan_mode: ScanMode[object, Sample],
    macro: Macro,
    scan_model: Model,
    helper: Helper,
    session: Session[Sample],
):
    distances = [1, 1]
    scan_height = 5
    dist = 1

    params = mocker.MagicMock()
    params.get = mocker.Mock(return_value="scan")
    params.get_float = mocker.Mock(return_value=scan_height)
    helper.prepare_scan_path = mocker.Mock(return_value=[MeshPoint(10, 10, False), MeshPoint(10.3, 10.3, True)])
    session.get_items = lambda: [Sample(time=0.0, frequency=dist) for _ in range(2)]
    scan_model.frequency_to_distance = mocker.Mock(side_effect=distances)

    finalize_spy = mocker.spy(helper, "finalize")

    macro.run(params)

    assert finalize_spy.mock_calls == [
        mocker.call(Position(0, 0, 0), [Position(10.3, 10.3, scan_height + scan_mode.probe_height - dist)])
    ]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import pytest
from typing_extensions import TypeAlias, override

from cartographer.printer_interface import HomingState, Mcu, Position, Toolhead
from cartographer.probe.scan_mode import Configuration, Model, ScanMode

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from cartographer.stream import Session


@dataclass
class Sample:
    time: float
    frequency: float
    position: Position | None = field(default_factory=lambda: Position(10, 10, 10))
    velocity: float | None = 5


Probe: TypeAlias = ScanMode[object, Sample]


class MockModel(Model):
    @property
    @override
    def name(self) -> str:
        return "test"

    actual_z_offset: float = 0

    @property
    @override
    def z_offset(self) -> float:
        return self.actual_z_offset

    @override
    def save_z_offset(self, new_offset: float) -> None:
        self.actual_z_offset = new_offset

    @override
    def distance_to_frequency(self, distance: float) -> float:
        return distance

    @override
    def frequency_to_distance(self, frequency: float) -> float:
        return frequency


class MockConfiguration(Configuration):
    x_offset: float = 0.0
    y_offset: float = 0.0
    move_speed: float = 42.0

    scan_samples: int = 10
    scan_mesh_runs: int = 1


@pytest.fixture
def config() -> MockConfiguration:
    return MockConfiguration()


@pytest.fixture
def model() -> Model:
    return MockModel()


@pytest.fixture
def homing_state(mocker: MockerFixture, probe: Probe) -> HomingState:
    mock = mocker.Mock(spec=HomingState, autospec=True)
    mock.endstops = [probe]
    return mock


@pytest.fixture
def probe(mcu: Mcu[object, Sample], toolhead: Toolhead, config: Configuration, model: Model) -> Probe:
    return ScanMode(mcu, toolhead, config, model=model)


def test_measures_distance(probe: Probe, session: Session[Sample]):
    session.get_items = lambda: [Sample(time=0.0, frequency=i + 1) for i in range(11)]

    distance = probe.measure_distance(skip_count=0)

    assert distance == 6


def test_skips_samples(probe: Probe, session: Session[Sample]):
    session.get_items = lambda: [Sample(time=0.0, frequency=i + 1) for i in range(11)]

    distance = probe.measure_distance(skip_count=4)

    assert distance == 8


def test_raises_error_when_model_is_missing(probe: Probe, model: Model):
    model.distance_to_frequency = lambda distance: 42

    frequency = probe.distance_to_frequency(1.0)

    assert frequency == 42


def test_probe_errors_when_not_homed(probe: Probe, toolhead: Toolhead):
    toolhead.is_homed = lambda axis: False

    with pytest.raises(RuntimeError):
        _ = probe.perform_probe()


def test_probe_returns_trigger_position(probe: Probe, toolhead: Toolhead, session: Session[Sample]):
    dist = 1
    z_pos = 2
    session.get_items = lambda: [Sample(time=0.0, frequency=dist) for _ in range(11)]
    toolhead.get_position = lambda: Position(0, 0, z_pos)

    trigger_pos = probe.perform_probe()

    assert trigger_pos == probe.probe_height + z_pos - dist


def test_probe_applies_axis_twist_compensation(probe: Probe, toolhead: Toolhead, session: Session[Sample]):
    dist = 1
    z_pos = 2
    z_comp = 0.5
    session.get_items = lambda: [Sample(time=0.0, frequency=dist) for _ in range(11)]
    toolhead.get_position = lambda: Position(0, 0, z_pos)
    toolhead.apply_axis_twist_compensation = lambda position: Position(position.x, position.y, z_comp)

    trigger_pos = probe.perform_probe()

    assert trigger_pos == z_comp


def test_probe_errors_outside_range(probe: Probe, session: Session[Sample], model: Model):
    session.get_items = lambda: [Sample(time=0.0, frequency=42) for _ in range(11)]
    model.frequency_to_distance = lambda frequency: float("inf")

    with pytest.raises(RuntimeError):
        _ = probe.perform_probe()


def test_do_nothing_when_not_homing(mocker: MockerFixture, probe: Probe, homing_state: HomingState):
    homed_position_spy = mocker.spy(homing_state, "set_z_homed_position")
    homing_state.is_homing_z = lambda: False
    probe.on_home_end(homing_state)
    assert homed_position_spy.call_count == 0


def test_do_nothing_when_endstop_not_homing(
    mocker: MockerFixture,
    probe: Probe,
    homing_state: HomingState,
):
    homed_position_spy = mocker.spy(homing_state, "set_z_homed_position")
    homing_state.is_homing_z = lambda: True
    homing_state.endstops = []
    probe.on_home_end(homing_state)
    assert homed_position_spy.call_count == 0


def test_scan_mode_sets_homed_position(
    mocker: MockerFixture,
    probe: Probe,
    homing_state: HomingState,
):
    homed_position_spy = mocker.spy(homing_state, "set_z_homed_position")
    probe.measure_distance = mocker.Mock(return_value=5)

    _ = probe.home_start(0)
    probe.on_home_end(homing_state)

    homed_position_spy.assert_called_once_with(5)


def test_endstop_is_triggered(mocker: MockerFixture, probe: Probe):
    probe.measure_distance = mocker.Mock(return_value=1)

    assert probe.query_is_triggered(0) is True


def test_endstop_is_not_triggered(mocker: MockerFixture, probe: Probe):
    probe.measure_distance = mocker.Mock(return_value=1)

    assert probe.query_is_triggered(0) is True


def test_save_z_offset_0(mocker: MockerFixture, probe: Probe, model: Model):
    save_model_offset_spy = mocker.spy(model, "save_z_offset")

    adjustment = 0.2
    new_offset = probe.offset.z - adjustment
    probe.save_z_offset(new_offset)

    assert probe.offset.z == 1.8
    assert save_model_offset_spy.mock_calls == [mocker.call(pytest.approx(-0.2))]  # pyright: ignore[reportUnknownMemberType]


def test_save_z_offset_existing(mocker: MockerFixture, probe: Probe, model: MockModel):
    model.actual_z_offset = 1.0
    save_model_offset_spy = mocker.spy(model, "save_z_offset")

    adjustment = 0.2
    new_offset = probe.offset.z - adjustment
    probe.save_z_offset(new_offset)

    assert probe.offset.z == 2.8
    assert save_model_offset_spy.mock_calls == [mocker.call(pytest.approx(0.8))]  # pyright: ignore[reportUnknownMemberType]



import logging
from typing import Literal, final

import numpy as np
from typing_extensions import override

from cartographer.printer_interface import Macro, MacroParams, S, Toolhead
from cartographer.probe.scan_mode import ScanMode

logger = logging.getLogger(__name__)


@final
class EstimateBacklashMacro(Macro[MacroParams]):
    name = "ESTIMATE_BACKLASH"
    description = "Do a series of moves to estimate backlash on the Z axis."

    def __init__(self, toolhead: Toolhead, scan: ScanMode[object, S]) -> None:
        self._scan = scan
        self._toolhead = toolhead

    @override
    def run(self, params: MacroParams) -> None:
        iterations = params.get_int("ITERATIONS", 10, minval=1)
        speed = 5
        height = 2
        delta = 0.2  # Smaller delta is better for detecting subtle backlash

        self._toolhead.move(z=height, speed=speed)
        samples: dict[Literal["up", "down"], list[float]] = {"up": [], "down": []}

        with self._scan.start_session():
            for _ in range(iterations):
                for direction in samples:
                    dir = 1 if direction == "up" else -1
                    self._toolhead.move(z=height + delta * dir, speed=speed)
                    self._toolhead.move(z=height, speed=speed)
                    self._toolhead.wait_moves()
                    dist = self._scan.measure_distance()
                    samples[direction].append(dist)

        global_mean = np.mean(samples["up"] + samples["down"])
        mean_up = np.mean(samples["up"]) - global_mean
        mean_down = np.mean(samples["down"]) - global_mean
        std_up = np.std(samples["up"])
        std_down = np.std(samples["down"])
        backlash = mean_down - mean_up  # Positive = down sits lower than up

        logger.info(
            """
            Backlash estimation results (normalized) over %d iterations:
            Mean up: %.5f mm
            Mean down: %.5f mm
            Std dev up: %.5f mm
            Std dev down: %.5f mm
            Estimated backlash (down - up): %.5f mm
            """,
            iterations,
            mean_up,
            mean_down,
            std_up,
            std_down,
            backlash,
        )
